/* abc.c: get allnet messages from ad, broadcast them to one interface */
/*        get allnet messages from the interface, forward them to ad */
/* abc stands for (A)llNet (B)road(C)ast */
/* single-threaded, uses select to check the pipe from ad and the interface */
/* must be run with supervisory privileges */
/* arguments are:
  - the fd number of the pipe from ad
  - the fd number of the pipe to ad
  - the interface name
 */
/* config file "abc" "interface-name" (e.g. ~/.allnet/abc/wlan0)
 * gives the maximum fraction of time the interface should be turned
 * on for allnet traffic.
 * if not found, the maximum fraction is 1 percent, i.e. 0.01
 * this fraction only applies to messages with priority 0.5 or less,
 * and on-time will be increased after we send messages with priority
 * above 0.5 or we receive messages with destination social distance
 * less than 2.
 *
 * for now (for testing, 2013/01/01), may be on all the time.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ifaddrs.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netpacket/packet.h>   /* sockaddr_ll */
#include <net/if.h>   		/* ifa_flags */

#include "packet.h"
#include "pipemsg.h"
#include "priority.h"
#include "util.h"

/* we don't know how big messages will be on the interface until we get them */
#define MAX_RECEIVE_BUFFER	ALLNET_MTU

/* similar to system(3), but more control over what gets printed */
static int my_system (char * command)
{
  pid_t pid = fork ();
  if (pid < 0) {
    perror ("fork");
    printf ("error forking for command '%s'\n", command);
    return -1;
  }
  if (pid == 0) {   /* child */
    int num_args = 1;
    char * argv [100];
    char * p = command;
    int found_blank = 0;
    argv [0] = command;
    while (*p != '\0') {
      if (found_blank) {
        if (*p != ' ') {
          argv [num_args] = p;
          num_args++;
          found_blank = 0;
        }
      } else if (*p == ' ') {
        found_blank = 1;
        *p = '\0';
      }
      p++;
    }
    argv [num_args] = NULL;
/*
    printf ("executing ");
    char ** debug_p = argv;
    while (*debug_p != NULL) {
      printf ("%s ", *debug_p);
      debug_p++;
    }
    printf ("\n");
*/
    dup2 (1, 2);   /* make stderr be a copy of stdout */
    execvp (argv [0], argv);
    perror ("execvp");
    exit (1);
  }
  /* parent */
  int status;
  do {
    waitpid (pid, &status, 0);
  } while (! WIFEXITED (status));
/*
  printf ("child process (%s) exited, status is %d\n",
          command, WEXITSTATUS (status));
*/
  return (WEXITSTATUS (status));
}

/* return 1 if successful, 0 otherwise */
/* return 2 if failed, but returned status matches wireless_status */
static int if_command (char * basic_command, char * interface,
                       int wireless_status, char * fail_wireless,
                       char * fail_other)
{
  static int printed_success = 0;
  int size = strlen (basic_command) + strlen (interface) + 1;
  char * command = malloc (size);
  if (command == NULL) {
    printf ("abc: unable to allocate %d bytes for command:\n", size);
    printf (basic_command, interface);
    return 0;
  }
  snprintf (command, size, basic_command, interface);
  int sys_result = my_system (command);
  int max_print_success = 0;
#ifdef DEBUG_PRINT
  max_print_success = 4;
#endif /* DEBUG_PRINT */
  if ((sys_result != 0) || (printed_success++ < max_print_success))
    printf ("abc: result of calling '%s' was %d\n", command, sys_result);
  if (sys_result != 0) {
    if (sys_result != -1)
      printf ("abc: program exit status for %s was %d\n",
              command, sys_result);
    if (sys_result != wireless_status) {
      if (fail_other != NULL)
        printf ("abc: call to '%s' failed, %s\n", command, fail_other);
      else
        printf ("abc: call to '%s' failed\n", command);
    } else {
      printf ("abc: call to '%s' failed, %s\n", command, fail_wireless);
      return 2;
    }
    return 0;
  }
  return 1;
}

/* returns 1 if successful, 2 if already up, 0 for failure */
static int wireless_up (char * interface)
{
#ifdef DEBUG_PRINT
  printf ("abc: opening interface %s\n", interface);
#endif /* DEBUG_PRINT */
/* need to execute the commands:
      sudo iw dev $if set type ibss
      sudo ifconfig $if up
      sudo iw dev $if ibss join allnet 2412
 */
  char * mess = "probably a wired or configured interface";
  if (geteuid () != 0)
    mess = "probably need to be root";
  int r = if_command ("iw dev %s set type ibss", interface, 240,
                      "wireless interface not available for ad-hoc mode",
                      mess);
  if (r == 0)
    return 0;
  if (r == 2) /* already up, no need to bring up the interface */
    return 2;
  /* continue with the other commands, which should succeed */
  if (! if_command ("ifconfig %s up", interface, 0, NULL, NULL))
    return 0;
  r = if_command ("iw dev %s ibss join allnet 2412", interface,
                  142, "allnet ad-hoc mode already set", "unknown problem");
  if (r == 0)
    return 0;
  if (r == 0)
    return 2;
  return 1;
}

/* returns 1 if successful, 0 for failure */
static int wireless_down (char * interface)
{
#ifdef DEBUG_PRINT
  printf ("taking down interface %s\n", interface);
#endif /* DEBUG_PRINT */
/* doesn't seem to be necessary or helpful
  if (! if_command ("iw dev %s set type managed", interface, NULL))
    return 0;
*/
  if (! if_command ("ifconfig %s down", interface, 0, NULL, NULL))
    return 0;
  return 1;
}

static void default_broadcast_address (struct sockaddr_ll * bc)
{
  bc->sll_family = AF_PACKET;
  bc->sll_protocol = ALLNET_WIFI_PROTOCOL;
  bc->sll_hatype = 1;   /* used? */
  bc->sll_pkttype = 0;  /* not used */
  bc->sll_halen = 6;
  bc->sll_addr [0] = 0xff;
  bc->sll_addr [1] = 0xff;
  bc->sll_addr [2] = 0xff;
  bc->sll_addr [3] = 0xff;
  bc->sll_addr [4] = 0xff;
  bc->sll_addr [5] = 0xff;
  printf ("set default broadcast address\n");
}

static void print_sll_addr (struct sockaddr_ll * a, char * desc)
{
  if (desc != NULL)
    printf ("%s: ", desc);
  if (a->sll_family != AF_PACKET) {
    printf ("unknown address family %d\n", a->sll_family);
    return;
  }
  printf ("proto %d, ha %d pkt %d halen %d ", a->sll_protocol, a->sll_hatype,
          a->sll_pkttype, a->sll_halen);
  int i;
  for (i = 0; i < a->sll_halen; i++) {
    if (i > 0) printf (":");
    printf ("%02x", a->sll_addr [0]);
  }
  if (desc != NULL)
    printf ("\n");
}

/* returns -1 if the interface is not found */
/* returns 0 if the interface is off, and 1 if it is on already */
/* if returning 0 or 1, fills in the socket and the address */
static int init_wireless (char * interface, int * sock,
                          struct sockaddr_ll * address, struct sockaddr_ll * bc)
{
  struct ifaddrs * ifa;
  if (getifaddrs (&ifa) != 0) {
    perror ("getifaddrs");
    exit (1);
  }
  struct ifaddrs * ifa_loop = ifa;
  while (ifa_loop != NULL) {
    if ((ifa_loop->ifa_addr->sa_family == AF_PACKET) &&
        (strcmp (ifa_loop->ifa_name, interface) == 0)) {
      struct timeval start;
      gettimeofday (&start, NULL);
      int is_up = wireless_up (interface);
      int in_use = (is_up == 2);
      if (is_up) {
        struct timeval midtime;
        gettimeofday (&midtime, NULL);
        long long mtime = delta_us (&midtime, &start);
        if (! in_use) {
          wireless_down (interface);
          struct timeval finish;
          gettimeofday (&finish, NULL);
          long long time = delta_us (&finish, &start);
          printf ("abc: %s is wireless, %lld.%03lld ms to turn on+off\n",
                  interface, time / 1000LL, time % 1000LL);
          printf ("  (%lld.%03lld ms to turn on)\n",
                  mtime / 1000LL, mtime % 1000LL);
        }
        /* create the socket and initialize the address */
        *sock = socket (AF_PACKET, SOCK_DGRAM, ALLNET_WIFI_PROTOCOL);
        *address = *((struct sockaddr_ll *) (ifa_loop->ifa_addr));
        if (ifa_loop->ifa_flags & IFF_BROADCAST)
          *bc = *((struct sockaddr_ll *) (ifa_loop->ifa_broadaddr));
        else if (ifa_loop->ifa_flags & IFF_POINTOPOINT)
          *bc = *((struct sockaddr_ll *) (ifa_loop->ifa_dstaddr));
        else
          default_broadcast_address (bc);
        bc->sll_protocol = ALLNET_WIFI_PROTOCOL;  /* otherwise not set */
        print_sll_addr (address, "interface address");
        print_sll_addr (bc,      "broadcast address");
        freeifaddrs (ifa);
        return in_use;
      }
    }
    ifa_loop = ifa_loop->ifa_next;
  }
  freeifaddrs (ifa);
  return -1;  /* interface not found */
}

static void send_beacon (int fd, unsigned char * dest, int nbits, int hops,
                         struct sockaddr * addr, socklen_t addrlen)
{
  static struct timeval send_next = {0, 0};
  struct timeval now;
  gettimeofday (&now, NULL);
  if (send_next.tv_sec == 0)  /* seed the random number generator */
    srandom (now.tv_sec);
  if (delta_us (&now, &send_next) == 0LL)   /* do not send yet */
    return;
  /* send the next beacon at a random time between 0.5s and 1.5s from now */
  set_time_random (&now, HALF_SECOND, ONE_SECOND + HALF_SECOND, &send_next);
/*
  printf ("%ld.%06ld sending beacon, next %ld.%06ld delta %lld.%06lld\n",
          now.tv_sec, now.tv_usec, send_next.tv_sec, send_next.tv_usec,
          delta_us (&send_next, &now) / US_PER_S,
          delta_us (&send_next, &now) % US_PER_S);
*/
  char buffer [ALLNET_MGMT_HEADER_SIZE];
  struct allnet_header_mgmt * hp = (struct allnet_header_mgmt *) buffer;
  hp->packet_type = ALLNET_TYPE_MGMT;
  hp->hops = 0;
  hp->max_hops = hops;
  hp->num_destination_bits = nbits;
  if (nbits > 0)
    memcpy (hp->destination, dest, (nbits + 7) / 8);
  hp->mgmt_type = ALLNET_MGMT_BEACON;
  if (sendto (fd, buffer, ALLNET_MGMT_HEADER_SIZE, MSG_DONTWAIT, addr, addrlen)
      < ALLNET_MGMT_HEADER_SIZE)
    perror ("beacon sendto");
}

static void main_loop (int rpipe, int wpipe, char * interface)
{
  int sockfd;
  struct sockaddr_ll if_address; /* the address of the interface */
  struct sockaddr  * if_sap = (struct sockaddr *) (&if_address);
  struct sockaddr_ll bc_address; /* broacast address of the interface */
  struct sockaddr  * bc_sap = (struct sockaddr *) (&bc_address);
  socklen_t addrlen = sizeof (struct sockaddr_ll);
  int in_use;                    /* if in use, no need to bring up and down */

  /* init sockfd */
  in_use = init_wireless (interface, &sockfd, &if_address, &bc_address);
  if (in_use < 0) {
    printf ("unable to bring up interface %s, for now aborting\n", interface);
    exit (1);
  }
  wireless_up (interface);
  printf ("wireless interface %s is up\n", interface);
  send_beacon (sockfd, NULL, 0, 1, bc_sap, addrlen);

  int fds [2];
  fds [0] = rpipe;
  fds [1] = sockfd;
  while (1) {
    /* int fd = next_available (fds, 2, PIPE_MESSAGE_WAIT_FOREVER); */
    int fd = next_available (fds, 2, 1000);
    if (fd == -1) {
      send_beacon (sockfd, NULL, 0, 1, bc_sap, addrlen);
      continue;   /* timeout */
    }
    if (fd == -1) {
      printf ("abc error: next_available returned -1\n");
      exit (1);
    }
    if (fd == sockfd) {
      /* printf ("expecting data on the socket\n"); */
      static char buf [MAX_RECEIVE_BUFFER];
      int r = recv (sockfd, buf, MAX_RECEIVE_BUFFER, MSG_DONTWAIT);
      /* printf ("got %d bytes on socket %d\n", r, sockfd); */
      if (r < 0)
        perror ("socket recv");
      if (r == 0)
        printf ("got 0 bytes on the socket\n");
      if (r <= 0)  /* something wrong?  socket closed?  who knows */
        break;
      /* print_buffer (buf, r); */
      if (r > MAX_RECEIVE_BUFFER) {
        printf ("got %d bytes, buffer is %d bytes\n", r, MAX_RECEIVE_BUFFER);
        r = MAX_RECEIVE_BUFFER;
      }
      if (! send_pipe_message (wpipe, buf, r, ONE_EIGHT)) {
        printf ("unable to send message to ad\n");
        break;
      }
    } else {  /* fd == rpipe */
      char * buffer;
      int priority;
      int dlen = receive_pipe_message (rpipe, &buffer, &priority);
      if (dlen <= 0) {
        printf ("error getting message from ad\n");
        break;
      }
/*
      printf ("ad message with %d bytes, priority ", dlen);
      print_fraction (priority, NULL);
      printf ("\n");
      print_buffer (buffer, dlen);
*/
      
      /* send the packet on the interface */
      int s = sendto (sockfd, buffer, dlen, MSG_DONTWAIT, bc_sap, addrlen);
      if (s < dlen)
        perror ("sendto");
      int debug = 0;
      if ((s < dlen) || (debug)) {
        printf ("sent %d bytes on interface %s, result %d\n",
                dlen, interface, s);
        /* print_sll_addr (&bc_address, NULL); */
      }
      if (s < dlen)
        break;
      /* printf ("abc main loop freeing buffer %p\n", buffer); */
      free (buffer);
    }
  }
  wireless_down (interface);
}

int main (int argc, char ** argv)
{
  if (argc != 4) {
    printf ("arguments must be a read pipe, a write pipe, and an interface\n");
    printf ("argc == %d\n", argc);
    return -1;
  }
  int rpipe = atoi (argv [1]);  /* read pipe */
  int wpipe = atoi (argv [2]);  /* write pipe */
  char * interface = argv [3];

  printf ("read pipe is fd %d, write pipe fd %d, interface is '%s'\n",
          rpipe, wpipe, interface);
  main_loop (rpipe, wpipe, interface);
  printf ("end of abc (%s) main thread\n", interface);
}
