/* acache.c: cache all data messages and respond to requests */
/* only one thread, listening on a pipe from ad, and responding
 * acache takes two arguments, the fd of a pipe from AD and of a pipe to AD
 */
/* for now, simply dcache all the packets, return them if they are in the
 * cache.  Later on, maybe provide persistent storage */
/* note that we never tell dcache that the storage is used, so dcache
 * eliminates messages in fifo order */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "packet.h"
#include "pipemsg.h"
#include "priority.h"
#include "util.h"
#include "dcache.h"
#include "sha.h"

/* #define CACHE_SIZE	1024 */
#define CACHE_SIZE	8  /* for testing */
struct cache_entry {
  char * message;
  int msize;
};

static struct cache_entry cache_storage [CACHE_SIZE];
static int cache_search = 0;

static void return_cache_entry (void * arg)
{
  struct cache_entry * cep = (struct cache_entry *) arg;
  free (cep->message);
  cep->msize = 0;   /* mark it as free */
}

static void save_packet (void * cache, char * message, int msize)
{
  struct allnet_header * hp = (struct allnet_header *) message;
  int i;
  for (i = 0; i < CACHE_SIZE; i++) {
    if (cache_storage [(i + cache_search) % CACHE_SIZE].msize == 0) {
      /* found a free slot -- since the dcache has CACHE_SIZE - 1 entries,
         there should always be a free slot. */
      struct cache_entry * cep = 
        cache_storage + ((i + cache_search) % CACHE_SIZE);
      cache_search = i;
      cep->message = message;
      cep->msize = msize;
      cache_add (cache, cep);  /* may call return_cache_entry */
      return;
    }
  }
  printf ("unable to find a free slot!!!  message not saved\n");
  free (message);
}

static int request_matches_packet (void * packet, void * cache_entry)
{
  struct cache_entry * cep = (struct cache_entry *) cache_entry;

  struct allnet_header_data * hp = (struct allnet_header_data *) packet;
  int nbits = hp->packet_id [0];

  struct allnet_header * chp = (struct allnet_header *) (cep->message);
  if (matches (hp->packet_id + 1, hp->packet_id [0],
               chp->destination,  chp->num_destination_bits))
    return nbits;
  return 0;
}

static void respond_to_packet (void * cache, char * message, int msize, int fd)
{
  struct allnet_header * hp = (struct allnet_header *) message;
  if (msize < ALLNET_DATA_HEADER_SIZE) {
    printf ("respond_to_packet: size %d, min %ld, ignoring request\n", msize,
            ALLNET_DATA_HEADER_SIZE);
    return;
  }
  struct cache_entry * cep = 
    cache_get_match (cache, request_matches_packet, message);
  if (cep != NULL)   /* send, no need to even check the return value */
    send_pipe_message (fd, cep->message, cep->msize, CACHE_RESPONSE_PRIORITY);
}

static int match_for_ack (void * packet_id, void * cache_entry)
{
  struct cache_entry * cep = (struct cache_entry *) cache_entry;
  if (cep->msize < ALLNET_DATA_HEADER_SIZE)
    return 0;
  struct allnet_header_data * hp = (struct allnet_header_data *) cep->message;
  if (memcmp (hp->packet_id, packet_id, PACKET_ID_SIZE) == 0)
    return 1;
  return 0;
}

static void ack_packet (void * cache, char * message, int msize)
{
  if (msize < ALLNET_DATA_HEADER_SIZE) {
    printf ("ack_packet: size %d, min %ld, ignoring ack\n", msize,
            ALLNET_DATA_HEADER_SIZE);
    return;
  }
  struct allnet_header_data * hp = (struct allnet_header_data *) message;
  char hash [SHA512_SIZE];
  sha512 (hp->packet_id, PACKET_ID_SIZE, hash);
  void * found;
  while ((found = cache_get_match (cache, match_for_ack, hash)) != NULL)
    cache_remove (cache, found);
}

void * main_loop (int rpipe, int wpipe)
{
  bzero (cache_storage, sizeof (cache_storage));
  void * cache = cache_init (CACHE_SIZE - 1, return_cache_entry);
  while (1) {
    char * message;
    int priority;
    int result = receive_pipe_message (rpipe, &message, &priority);
    if (result <= 0) {
      printf ("acache ad pipe %d closed, result %d\n", rpipe, result);
      break;
    }
    /* message from ad: save, respond, or ignore */
    struct allnet_header * hp = (struct allnet_header *) message;
    if ((hp->packet_type == ALLNET_TYPE_DATA) ||
        (hp->packet_type == ALLNET_TYPE_KEY_XCHG) ||
        (hp->packet_type == ALLNET_TYPE_CLEAR)) {  /* save it */
      save_packet (cache, message, result);
      printf ("saved packet of type %d\n", hp->packet_type);
    } else if (hp->packet_type == ALLNET_TYPE_DATA_REQ) { /* respond */
      respond_to_packet (cache, message, result, wpipe);
      free (message);
    } else if (hp->packet_type == ALLNET_TYPE_DATA_ACK) { /* erase if have */
      ack_packet (cache, message, result);
      free (message);
    } else {                                                /* ignore */
      printf ("acache ignoring packet of type %d\n", hp->packet_type);
      free (message);
    }
  }
}

int main (int argc, char ** argv)
{
  if (argc != 3) {
    printf ("arguments must be a read and a write pipe\n");
    return -1;
  }
/*
  printf ("in acache, args are ");
  printf ("'%s %s %s'\n", argv [0], argv [1], argv [2]);
*/
  int rpipe = atoi (argv [1]);
  int wpipe = atoi (argv [2]);
  /* printf ("read pipe is fd %d, write pipe is fd %d\n", rpipe, wpipe); */

  main_loop (rpipe, wpipe);
  printf ("end of acache\n");
}
