/* ad.c: main allnet daemon to forward allnet messages */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include "packet.h"
#include "pipemsg.h"
#include "social.h"
#include "priority.h"

#define PROCESS_PACKET_DROP	0
#define PROCESS_PACKET_LOCAL	1
#define PROCESS_PACKET_MGMT	2
#define PROCESS_PACKET_ALL	4

/* return 0 to drop the packet (do nothing), 1 to process as a request packet,
 * 2 to forward only to local destinations, and 3 to forward everywhere */
/* if returning 3, fills in priority */
static int process_packet (char * packet, int size, int is_local,
                           struct social_info * soc, int * priority)
{
  if (size < ALLNET_BASIC_HEADER_SIZE) {
    printf ("received a packet with %d bytes, %zd required\n",
            size, ALLNET_BASIC_HEADER_SIZE);
    return PROCESS_PACKET_DROP;
  }
/* skip the hop count in the hash, since it changes at each hop */
#define HEADER_SKIP	2
  if (record_packet (packet + HEADER_SKIP, size - HEADER_SKIP)) {
#undef HEADER_SKIP
    printf ("packet received previously, dropping\n");
    return PROCESS_PACKET_DROP;
  }

 /* received a message with a header */
  struct allnet_header * ah = (struct allnet_header *) packet;

  if (ah->packet_type == ALLNET_TYPE_MGMT)      /* AllNet management */
    return PROCESS_PACKET_MGMT;

  if (is_local)
    return PROCESS_PACKET_ALL;

  if (ah->hops == 255) {  /* something wrong, cannot increment */
    printf ("error: cannot forward packet with 255 hops already\n");
    return PROCESS_PACKET_LOCAL;
  }

  /* before forwarding, increment the number of hops seen */
  ah->hops++;

  if (ah->hops >= ah->max_hops)
  /* no matter what it is, only forward locally, i.e. to alocal and acache */
    return PROCESS_PACKET_LOCAL;

  /* compute a forwarding priority for non-local packets */
  int social_distance =
    social_connection (soc, ah->destination, ah->num_destination_bits);
  int rate_fraction =
    track_rate (ah->destination, ah->num_destination_bits, size);
  *priority = compute_priority (is_local, size, ah->num_destination_bits,
                                ah->hops, ah->max_hops - ah->hops,
                                social_distance, rate_fraction);

  /* send each of the packets, with its priority, to each of the pipes */
  return PROCESS_PACKET_ALL;
};

static void send_all (char * packet, int psize, int priority,
                      int * write_pipes, int nwrite)
{
  printf ("send_all sending to %d pipes\n", nwrite);
  int i;
  for (i = 0; i < nwrite; i++)
    if (! send_pipe_message (write_pipes [i], packet, psize, priority))
      printf ("write_pipes [%d] = %d is no longer valid\n", i, write_pipes [i]);
}

/* runs forever, and only returns in case of error. */
/* the first read_pipe and the first write_pipe are from/to alocal.
 * the second read_pipe and write_pipe are from/to acache
 * the third read_pipe and write_pipe are from/to aip
 * there may or may not be more pipes, but they should generally be the
 * same number, even though the code only explicitly refers to the first
 * three and doesn't require the same number of read and write pipes
 */
static void main_loop (int * read_pipes, int nread,
                       int * write_pipes, int nwrite,
                       int update_seconds, int max_social_bytes)
{
  add_pipes (read_pipes, nread);
  struct social_info * soc = init_social (max_social_bytes);
  time_t next_update = update_social (soc, update_seconds);

  while (1) {
    /* read messages from each of the pipes */
    char * packet;
    int from_pipe;
    int priority = 0;   /* incoming priorities ignored unless from local */
    int psize = receive_pipe_message_any (PIPE_MESSAGE_WAIT_FOREVER,
                                          &packet, -1, &from_pipe, &priority);
printf ("ad received %d, fd %d\n", psize, from_pipe);
    if (psize <= 0) { /* for now exit */
      printf ("error: received %d from receive_pipe_message_any, pipe %d\n",
              psize, from_pipe);
      printf (" exiting\n");
      exit (1);
    }
    /* packets generated by alocal and acache are local */
    int is_local = ((from_pipe == read_pipes [0]) ||
                    (from_pipe == read_pipes [1]));
    int p = process_packet (packet, psize, is_local, soc, &priority);
    switch (p) {
    case PROCESS_PACKET_ALL:
      send_all (packet, psize, priority, write_pipes, nwrite);
      break;
    /* all the rest are not forwarded, so priority does not matter */
    case PROCESS_PACKET_LOCAL:   /* send only to alocal and acache */ 
      send_all (packet, psize, 0, write_pipes, 2);
      break;
    case PROCESS_PACKET_MGMT:    /* send only to aip */
      send_all (packet, psize, priority, write_pipes + 2, 1);
      break;
    case PROCESS_PACKET_DROP:    /* do not forward */
      /* do nothing */
      break;
    }
    free (packet);  /* was allocated by receive_pipe_message_any */

    /* about once every next_update seconds, re-read social connections */
    if (time (NULL) >= next_update)
      next_update = update_social (soc, update_seconds);
  }
}

/* arguments are: the number of pipes, then pairs of read and write file
 * file descriptors (ints) for each pipe, from/to alocal, acache, aip.
 * any additional pipes will again be pairs from/to each abc.
 */
int main (int argc, char ** argv)
{
  if (argc < 2) {
    printf ("need to have at least the number of read and write pipes\n");
    return -1;
  }
  int npipes = atoi (argv [1]);
  if (npipes < 3) {
    printf ("%d pipes, at least 3 needed\n", npipes);
    return -1;
  }
  if (argc != 2 * npipes + 2) {
    printf ("%d arguments, expected 2 + %d for %d pipes\n",
            argc, 2 * npipes, npipes);
    return -1;
  }
  if (argc < 7) {
    printf ("need to have at least 3 each read and write pipes\n");
    return -1;
  }
  /* allocate both read and write pipes at once, then point write_pipes
   * to the middle of the allocated array
   */
  int * read_pipes  = malloc (sizeof (int) * npipes * 2);
  if (read_pipes == NULL) {
    printf ("allocation error in ad main\n");
    return -1;
  }
  int * write_pipes = read_pipes + npipes;

  int i;
  for (i = 0; i < npipes; i++) {
    read_pipes  [i] = atoi (argv [2 + 2 * i    ]);
    write_pipes [i] = atoi (argv [2 + 2 * i + 1]);
  }
  for (i = 0; i < npipes; i++)
    printf ("read_pipes [%d] = %d\n", i, read_pipes [i]);
  for (i = 0; i < npipes; i++)
    printf ("write_pipes [%d] = %d\n", i, write_pipes [i]);
  main_loop (read_pipes, npipes, write_pipes, npipes, 30, 30000);
  printf ("ad error: main loop returned, exiting\n");
}

