/* aip.c: get allnet messages from ad, send them to DHT and known hosts */
/*        get allnet messages from the internet, forward them to ad */
/* aip stands for A(llNet) IP interface */
/* main thread uses select to check the pipe from ad and the sockets */
/* secondary threads:
 * - listen and open TCP connections
 * - listen on the unix socket for allnet-destination-address to IP mappings
 */
/* arguments are:
  - the fd number of the pipe from ad
  - the fd number of the pipe to ad
  - the name of the Unix socket
 */
/* config file "aip" "speed" (e.g. ~/.allnet/aip/speed)
 * gives the maximum speed to send over the internet, in bytes/second
 * this speed limit only applies to messages with priority 0.5 or less,
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <pthread.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "packet.h"
#include "pipemsg.h"
#include "priority.h"
#include "util.h"
#include "listen.h"
#include "dcache.h"

/* UDPv4 messages are limited to less than 2^16 bytes */
#define MAX_RECEIVE_BUFFER	ALLNET_MAX_UDP_SIZE

struct addr_info {
  unsigned char destination [DESTINATION_SIZE];
  unsigned char nbits;    /* how many bits of the destination are given */
  unsigned char ip_version;   /* 4 or 6 */
  unsigned short port;
  struct in6_addr ip;         /* for ip4, only 4 bytes are filled/sent */
};

static int init_unix_socket (char * addr_socket_name)
{
  int addr_socket = socket (AF_UNIX, SOCK_DGRAM, 0);
  if (addr_socket < 0) {
    perror ("af_unix socket");
    exit (1);    /* not sure what this error would mean */
  }
  struct sockaddr_un sun;
  sun.sun_family = AF_UNIX;
  snprintf (sun.sun_path, sizeof (sun.sun_path), "%s", addr_socket_name);
  if (bind (addr_socket, (struct sockaddr *) (&sun), sizeof (sun)) < 0) {
    perror ("af_unix bind");
    printf ("unable to bind to unix socket ");
    print_sockaddr ((struct sockaddr *) (&sun), sizeof (sun));
    printf ("\nmake sure no other aip is running and delete %s\n",
            addr_socket_name);
    exit (1);
  }
  return addr_socket;
}

struct receive_arg {
  char * socket_name;
  void * cache;
};

static void * receive_addrs (void * arg)
{
  struct receive_arg * ra = (struct receive_arg *) arg;
  int addr_socket = init_unix_socket (ra->socket_name);
  printf ("receive_addrs, socket is %d\n", addr_socket);

  while (1) {
    int size = sizeof (struct addr_info);
    struct addr_info * ai = malloc_or_fail (size, "receive_addrs");
    int n = recv (addr_socket, (char *) ai, size, 0);
    if (n < 0)
      perror ("recv");
    if (n <= 0) {
      printf ("error: address socket %d closed, thread exiting\n", addr_socket);
      free (ai);
      return NULL;
    }
    if ((ai->ip_version == 4) && (n + 12 != size))
      printf ("ip version 4, expected size %d, got size %d\n", size - 12, n);
    if ((ai->ip_version == 6) && (n != size))
      printf ("ip version 6, expected size %d, got size %d\n", size, n);

    printf ("got IP%d ", ai->ip_version);
    if (ai->ip_version == 4)
      printf ("%d.%d.%d.%d",
              ai->ip.s6_addr [0], ai->ip.s6_addr [1], ai->ip.s6_addr [2],
              ai->ip.s6_addr [3]);
    else
      printf ("%d:%d:%d:%d:%d:%d:%d:%d:%d:%d:%d:%d:%d:%d:%d:%d",
              ai->ip.s6_addr [ 0], ai->ip.s6_addr [ 1], ai->ip.s6_addr [ 2],
              ai->ip.s6_addr [ 3], ai->ip.s6_addr [ 4], ai->ip.s6_addr [ 5],
              ai->ip.s6_addr [ 6], ai->ip.s6_addr [ 7], ai->ip.s6_addr [ 8],
              ai->ip.s6_addr [ 9], ai->ip.s6_addr [10], ai->ip.s6_addr [11],
              ai->ip.s6_addr [12], ai->ip.s6_addr [13], ai->ip.s6_addr [14],
              ai->ip.s6_addr [15]);
    printf (" port %d/0x%x for: ", ntohs (ai->port), ntohs (ai->port));
    print_buffer (ai->destination, (ai->nbits + 7) / 8, NULL, 5, 1);

    int id = (ai->nbits & 0xff)           << 24 |
             (ai->destination [0] & 0xff) << 16 |
             (ai->destination [1] & 0xff) <<  8 |
             (ai->destination [2] & 0xff);
    cache_add (ra->cache, ai);   /* if already in cache, records usage */
  }
}

static void addone (void * arg, void * data)
{
  int * countp = (int *) arg;
  (*countp) ++;
}

/* return 0 in case of no match, 1 if the first min(xbits, ybits) match */
static int matches_old (unsigned char * x, unsigned char * y, int bits)
{
  return matches (x, bits, y, bits);
}

struct find_arg {
  /* arguments */
  unsigned char destination [DESTINATION_SIZE];
  unsigned char nbits;    /* how many bits of the destination are given */
  unsigned int max_dests; /* this many sockaddrs are allocated */
  /* intermediate value -- set to zero before calling */
  unsigned int num_found;
  /* results */
  struct addr_info results [0];  /* actually max_dests */
};

static void print_addr_info (struct addr_info * ai)
{
  printf (" %p (%d) ", ai, ai->nbits);
  print_buffer (ai->destination, (ai->nbits + 7) / 8, NULL, 4, 0);
  printf (", v %d, port %d, addr ", ai->ip_version, ntohs (ai->port));
  print_buffer ((char *) &(ai->ip), 16, NULL, 16, 1);
}
static void print_find_arg (struct find_arg * fa)
{
  printf ("find_arg dest (%d) ", fa->nbits);
  print_buffer (fa->destination, (fa->nbits + 7) / 8, NULL, 4, 0);
  printf (", max %d, found %d\n", fa->max_dests, fa->num_found);
  int i;
  for (i = 0; i < fa->num_found; i++) {
    printf (" %2d: ", i);
    print_addr_info (fa->results + i);
  }
}

static void find_destinations (void * a, void * data)
{
  struct find_arg * fa = (struct find_arg *) a;
  struct addr_info * ai = (struct addr_info *) data;
  printf ("find_destinations: \n");
  print_addr_info (ai);
  print_find_arg (fa);
  int i, j;
  int min_bits = ((ai->nbits > fa->nbits) ? fa->nbits : ai->nbits);
  for (i = 0; i < fa->num_found; i++) {
    /* the minimum number of bits between the cache entry and the search */
    printf (" %d: min_bits %d\n", i, min_bits);
    /* if we match, will we do better than the current entry?
     * if so, try to match */
    if ((fa->results [i].nbits < min_bits) &&
        (matches_old (fa->destination, ai->destination, min_bits))) {
      printf ("match found at %d\n", i);
    /* shift this and later entries out of the way to make room for this */
      for (j = fa->max_dests; j > i; j--)
        fa->results [j] = fa->results [j - 1];
      fa->results [i] = *ai;
      fa->results [i].nbits = min_bits;
      if (fa->num_found < fa->max_dests)
        fa->num_found++;
      print_find_arg (fa);
      return;  /* done inserting this item */
    }
  }
  if ((fa->num_found < fa->max_dests) &&  /* try to add at the end */
      (matches_old (fa->destination, ai->destination, min_bits))) {
    fa->results [fa->num_found] = *ai;
    fa->results [fa->num_found].nbits = min_bits;
    fa->num_found++;
    print_find_arg (fa);
  }
}

/* mallocs and sets result to an n-element array of addr_infos that are the
 * best matches for the given destination */
/* returns the actual number of destinations found, or 0 */
static int top_destinations_old (void * addr_cache, int n, unsigned char * dest,
                             int nbits, struct sockaddr_in6 ** result)
{
  if ((nbits / 8) > DESTINATION_SIZE)
    nbits = DESTINATION_SIZE * 8;
  *result = NULL;  /* in case we return 0 */
  int max = 0;
  cache_map (addr_cache, addone, &max);  /* find out how many there are */
  if (n > max) n = max;
  printf ("cache has %d addresses, searching for %d\n", max, n);
  if (n <= 0)
    return 0;
  int size = n * sizeof (struct sockaddr_in6 *);
  int fasize = sizeof (struct find_arg) + n * sizeof (struct addr_info);
  struct sockaddr_in6 * r = malloc (size);
  if (r == NULL) {
    printf ("unable to allocate %d bytes for %d destinations\n", size, n);
    return 0;
  }
  struct find_arg * fap = malloc (fasize);
  if (fap == NULL) {
    printf ("unable to allocate %d bytes for %d matches\n", fasize, n);
    free (r);
    return 0;
  }
  int i;
  memcpy (fap->destination, dest, (nbits + 7) / 8);
  fap->nbits = nbits;
  fap->max_dests = n;
  fap->num_found = 0;
  for (i = 0; i < n; i++)
    fap->results [i].ip_version = 0;
  cache_map (addr_cache, find_destinations, fap);
  int found = 0;
  for (i = 0; i < n; i++) {
    int debug_found = found;
    struct sockaddr_in  * sin  = (struct sockaddr_in  *) (r + found);
    struct sockaddr_in6 * sin6 =                         (r + found);
    if (fap->results [i].ip_version == 6) {
      sin6->sin6_family = AF_INET6;
      memcpy (&(sin6->sin6_addr), &(fap->results [i].ip), 16);
      sin6->sin6_port = fap->results [i].port;
      found++;
    } else if (fap->results [i].ip_version == 4) {
      sin->sin_family = AF_INET;
      memcpy (&(sin->sin_addr), &(fap->results [i].ip), 4);
      sin->sin_port = fap->results [i].port;
      found++;
    } /* else, not found */
    if (debug_found != found) {
      printf ("found address %d: ", debug_found);
      print_sockaddr ((struct sockaddr *) (r + debug_found), 0);
      printf ("\n");
    }
  }
  if (found != fap->num_found)
    printf ("top_destinations found %d != %d\n", found, fap->num_found);
  printf ("found %d top_destinations\n", found);
  *result = r;
  return found;
}

/* should return nonzero for a matching entry (higher values for a
 * better match), and 0 for no match */

struct match_arg {
  /* arguments */
  unsigned char destination [DESTINATION_SIZE];
  unsigned char nbits;    /* how many bits of the destination are given */
};

/* return 0 in case of no match, otherwise the number of matching bits */
static int matching_bits (void * a, void * data)
{
  struct match_arg * ma = (struct match_arg *) a;
  struct addr_info * ai = (struct addr_info *) data;
  if (! matches (ma->destination, ma->nbits, ai->destination, ai->nbits))
    return 0;
  /* else, they are identical */
  int result = DESTINATION_SIZE * 8;
  if (result > MAX_MATCH)
    result = MAX_MATCH;
  if (result > ma->nbits)
    result = ma->nbits;
  if (result > ai->nbits)
    result = ai->nbits;
  return result;
}

/* mallocs and sets result to an n-element array of addr_infos that are the
 * best matches for the given destination */
/* returns the actual number of destinations found, or 0 */
static int top_destinations (void * addr_cache, int n, unsigned char * dest,
                             int nbits, struct sockaddr_in6 ** result)
{
  void ** matches;
  struct match_arg dest_arg;
  memcpy (dest_arg.destination, dest, (nbits + 7) / 8); 
  dest_arg.nbits = nbits;
  int num_matches = cache_all_matches (addr_cache, matching_bits, &dest_arg,
                                       &matches);
  printf ("top_destination (");
  print_buffer (dest, (nbits + 7) / 8, NULL, 100, 0);
  printf (" (%d), %d)\n", nbits, num_matches);
  struct sockaddr_in6 * new =
    malloc_or_fail (sizeof (struct sockaddr_in6) * n, "top_destinations");
  if (num_matches > n)
    num_matches = n;    /* only return the first n matches */
  int i;
  for (i = 0; i < num_matches; i++) {
    struct sockaddr_in  * sin  = (struct sockaddr_in  *) (new + i);
    struct sockaddr_in6 * sin6 =                         (new + i);
    struct addr_info * ai = ((struct addr_info *) (matches [i]));
print_addr_info (ai);
    if (ai->ip_version == 6) {
      sin6->sin6_family = AF_INET6;
      memcpy (&(sin6->sin6_addr), &(ai->ip), 16);
      sin6->sin6_port = ai->port;
    } else if (ai->ip_version == 4) {
      sin->sin_family = AF_INET;
      memcpy (&(sin->sin_addr), &(ai->ip), 4);
      sin->sin_port = ai->port;
    } else {   /* not found */
      printf ("coding error: match %d of %d has version %d\n", i, num_matches, 
              ai->ip_version);
      printf ("sizeof (ip) is %ld\n", sizeof (ai->ip));
      exit (1);
    }
  }
  *result = new;
  return num_matches;
}

static int is_sockaddr (void * arg1, void * arg2)
{
  struct sockaddr_in6 * a1 = (struct sockaddr_in6 *) arg1;
  struct sockaddr_in6 * a2 = (struct sockaddr_in6 *) arg2;
  if (a1->sin6_family != a2->sin6_family)
    return 0;
  if (a1->sin6_family == AF_INET6) {
    if ((a1->sin6_port == a2->sin6_port) &&
        (memcmp (a1->sin6_addr.s6_addr, a2->sin6_addr.s6_addr,
                 sizeof (a1->sin6_addr)) == 0))
      return 1;
    return 0;
  }
  if (a1->sin6_family == AF_INET) {
    struct sockaddr_in * sin1 = (struct sockaddr_in  *) arg1;
    struct sockaddr_in * sin2 = (struct sockaddr_in  *) arg2;
    if ((sin1->sin_port == sin2->sin_port) &&
        (sin1->sin_addr.s_addr == sin2->sin_addr.s_addr))
      return 1;
    return 0;
  }
  printf ("is_sockaddr: unknown address family %d\n", a1->sin6_family);
  return 0;
}

/* save the IP address of the sender, unless it is already there */
static void add_sockaddr_to_cache (void * cache, struct sockaddr * addr)
{
  print_sockaddr (addr, sizeof (struct sockaddr_in6));
  void * found = cache_get_match (cache, is_sockaddr, addr);
  if (found != NULL) {  /* found */
    printf (" (already in cache)");
    cache_record_usage (cache, found); /* found, addr are different pointers */
    free (addr);   /* we will not save it, so free it */
    return;
  }

  /* else, add to cache */
  cache_add (cache, addr);
}

static int receive_udp (int fd, char ** message, void * cache)
{
  printf ("ready to receive UDP packet on fd %d\n", fd);
  *message = malloc (ALLNET_MTU);
  socklen_t size = sizeof (struct sockaddr_in6);
  struct sockaddr * ap = malloc (size);
  if (*message == NULL)
    printf ("unable to allocate %d bytes for receive_udp in aip", ALLNET_MTU);
  if (ap == NULL)
    printf ("unable to allocate %d bytes for ap in aip", size);
  if ((*message == NULL) || (ap == NULL))
    return -1;
  int result = recvfrom (fd, *message, ALLNET_MTU, MSG_DONTWAIT, ap, &size);
  if (result < 0) {
    if ((errno != EAGAIN) && (errno != EWOULDBLOCK))
      perror ("recvfrom");
    free (*message);
    free (ap);
    *message = NULL;
    if ((errno != EAGAIN) && (errno != EWOULDBLOCK))
      return -1;
    return 0;   /* EAGAIN or EWOULDBLOCK, no message ready at this time */
  }

  /* got a packet, save the IP address of the sender */
  printf ("got %d udp bytes, saving sender ", result);
  print_sockaddr (ap, size);
  add_sockaddr_to_cache (cache, ap);
  printf ("\n");
  return result;
}

static void send_udp (int udp, char * message, int msize, struct sockaddr * sa)
{
  socklen_t addr_len = sizeof (struct sockaddr_in6);
  if (sa->sa_family == AF_INET)
    addr_len = sizeof (struct sockaddr_in);
  int s = sendto (udp, message, msize, 0, sa, addr_len);
  if (s < msize) {
    perror ("sendto");
    printf ("error sending %d (sent %d) on udp %d to ", msize, s, udp);
    print_sockaddr (sa, 0);
    printf ("\n");
  } else {
    printf ("sent %d bytes to ", msize);
    print_sockaddr (sa, 0);
    printf ("\n");
  }
}

static void forward_message (int * fds, int num_fds, int udp, void * udp_cache,
                             void * addr_cache, char * message, int msize,
                             int priority, int max_send)
{
  int i;
  /* first send up to about half to the UDPs for which we have translations,
   * then the rest to a random permutation of other udps we have heard from 
   * and tcps we are connected to */
  struct allnet_header * hp = (struct allnet_header *) message;
  struct sockaddr_in6 * destinations;
  int max_translations = max_send / 2 + 1;
  int translations =
    top_destinations (addr_cache, max_translations, hp->destination,
                      hp->num_destination_bits, &destinations);
  for (i = 0; i < translations; i++)  /* send here first */
    send_udp (udp, message, msize, (struct sockaddr *) (destinations + i));
  if (translations > 0) free (destinations);
  printf ("forwarded to %d mappings\n", translations);

  int sent_fds = 0;
  int sent_udps = 0;
  int remaining = max_send - translations;
#define FORWARDING_UDPS	100
  void * udps [FORWARDING_UDPS];
  int nudps = cache_random (udp_cache, FORWARDING_UDPS, udps);
#undef FORWARDING_UDPS
  int size = num_fds + nudps;
  if (size > 0) {
    int * random_selection = random_permute (size);
    for (i = 0; i < size && i < remaining; i++) {
      int index = random_selection [i];
      if (index < num_fds) {
        if (! send_pipe_message (fds [index], message, msize, ONE_HALF))
          printf ("aip error sending to socket %d at %d\n", fds [index], i);
        sent_fds++;
      } else {
        /* udp cache contains sockaddrs */
        send_udp (udp, message, msize,
                  (struct sockaddr *) (udps [index - num_fds]));
        sent_udps++;
      }
    }
    free (random_selection);
  }
  printf ("forwarded to %d TCP and %d UDP\n", sent_fds, sent_udps);
}

static int udp_socket ()
{
  int udp = socket (AF_INET6, SOCK_DGRAM, 0);
  if (udp < 0) {
    perror ("main loop socket");
    exit (1);
  }
  struct sockaddr_storage address;
  struct sockaddr     * ap  = (struct sockaddr     *) &address;
/*  struct sockaddr_in  * ap4 = (struct sockaddr_in  *) ap; */
  struct sockaddr_in6 * ap6 = (struct sockaddr_in6 *) ap;
  int addr_size = sizeof (address);

  memset (&address, 0, addr_size);
  ap6->sin6_family = AF_INET6;
  memcpy (&(ap6->sin6_addr), &(in6addr_any), sizeof (ap6->sin6_addr));
  ap6->sin6_port = ALLNET_PORT;
  if (bind (udp, ap, addr_size) < 0) {
    perror ("bind");
    printf ("aip unable to bind to UDP %d/%x, probably already running\n",
            ntohs (ALLNET_PORT), ntohs (ALLNET_PORT));
    exit (1);
  }
  printf ("opened UDP socket %d\n", udp);
  return udp;
}

static void main_loop (int rpipe, int wpipe, struct listen_info * info,
                       void * addr_cache)
{
  int udp = udp_socket ();
  void * udp_cache = NULL;
  udp_cache = cache_init (128, free);
  while (1) {
    int fd = 0;
    int priority;
    char * message;
    int result = receive_pipe_message_any (1000, &message, udp, &fd, &priority);
    if ((result == 0) && (fd == udp)) {
      result = receive_udp (fd, &message, udp_cache);
      priority = EPSILON;
    }
    if (result < 0) {
      if ((fd == rpipe) || (fd == udp)) {
        printf ("aip %s %d closed\n",
                ((fd == rpipe) ? "ad pipe" : "udp socket"), fd);
        break;  /* exit the loop and the program */
      }
      printf ("aip: error on file descriptor %d, closing\n", fd);
      listen_remove_fd (info, fd); /* remove from data structures */
      close (fd);       /* remove from kernel */
    } else if (result > 0) {
      if (fd == rpipe) {    /* message from ad, send to IP neighbors */
        forward_message (info->fds + 1, info->num_fds - 1, udp, udp_cache,
                         addr_cache, message, result, priority, 4);
      } else {              /* message from a client, send to ad */
        /* often ad will just compute the priority and send it back to us */
        if (! send_pipe_message (wpipe, message, result, EPSILON)) {
          printf ("aip error sending to ad pipe %d\n", wpipe);
          break;
        }
        listen_record_usage (info, fd);   /* this fd was used */
      }
      free (message);
    }   /* else result is zero, timed out, try again */
  }
}

int main (int argc, char ** argv)
{
  if (argc != 4) {
    printf ("aip: arguments are read pipe from ad and write pipe to ad,\n");
    printf (" and a unix domain socket for address info (argc == 4)\n");
    printf (" but argc == %d\n", argc);
    return -1;
  }

  int rpipe = atoi (argv [1]);  /* read pipe */
  int wpipe = atoi (argv [2]);  /* write pipe */
  char * addr_socket_name = argv [3];

  printf ("read pipe is fd %d, write pipe fd %d, socket %s\n",
          rpipe, wpipe, addr_socket_name);

  pthread_t addr_thread;
  struct receive_arg ra;
  ra.socket_name = addr_socket_name;
  ra.cache = cache_init (128, free);
  if (pthread_create (&addr_thread, NULL, receive_addrs, &ra) != 0) {
    perror ("pthread_create/addrs");
    return 1;
  }
  struct listen_info info;
  listen_init_info (&info, 256, "aip", ALLNET_PORT, 0, 1);

  listen_add_fd (&info, rpipe);

  srandom (time (NULL));
  main_loop (rpipe, wpipe, &info, ra.cache);

  printf ("end of aip main thread, deleting %s\n", addr_socket_name);
  if (unlink (addr_socket_name) < 0)
    perror ("unlink");
}
