/* astart.c: start all the processes used by the allnet daemon */
/* takes as argument the interface(s) on which to start
 * sending and receiving broadcast packets */
/* in the future this may be automated or from config file */
/* (since the config file should tell us the bandwidth on each interface) */
/* alternately, if the single program is named astop or the single argument
 * is "stop", stops running ad, assuming the pids are in
 * /var/run/allnet-pids or /tmp/allnet-pids */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <signal.h>

#include "util.h"

static void init_pipes (int * pipes, int num_pipes)
{
  int i;
  for (i = 0; i < num_pipes; i++) {
    if (pipe (pipes + i * 2) < 0) {
      perror ("pipe");
      printf ("error creating pipe set %d\n", i);
      exit (1);
    }
  }
}

static char * itoa (int n)
{
  char * result = malloc (12);  /* plenty of bytes, easier than being exact */
  snprintf (result, 10, "%d", n);
  return result;
}

static void print_pid (int fd, int pid)
{
  char * buffer = itoa (pid);
  int len = strlen (buffer);
  if (write (fd, buffer, len) != len)
    perror ("pid file write");
  buffer [0] = '\n';
  if (write (fd, buffer, 1) != 1)
    perror ("pid file newline write");
  free (buffer);
}

static void my_exec2 (char * program, int * pipes, int fd)
{
  int pid = fork ();
  if (pid == 0) {
    char * args [4];
    args [0] = program;
    args [1] = itoa (pipes [0]);
    args [2] = itoa (pipes [3]);
    args [3] = NULL;
    close (pipes [1]);
    close (pipes [2]);
    printf ("calling %s %s %s\n", args [0], args [1], args [2]);
    execv (program, args);
    printf ("error executing %s\n", program);
  } else {  /* parent, close the child pipes */
    print_pid (fd, pid);
    close (pipes [0]);
    close (pipes [3]);
    printf ("parent called %s %d %d, closed %d %d\n",
            program, pipes [0], pipes [3], pipes [0], pipes [3]);
  }
}

static void my_exec3 (char * program, int * pipes, char * extra, int fd)
{
  int pid = fork ();
  if (pid == 0) {
    char * args [5];
    args [0] = program;
    args [1] = itoa (pipes [0]);
    args [2] = itoa (pipes [3]);
    args [3] = extra;
    args [4] = NULL;
    close (pipes [1]);
    close (pipes [2]);
    printf ("calling %s %s %s %s\n", args [0], args [1], args [2], args [3]);
    execv (program, args);
    printf ("error executing %s\n", program);
  } else {  /* parent, close the child pipes */
    print_pid (fd, pid);
    close (pipes [0]);
    close (pipes [3]);
    printf ("parent called %s %d %d %s, closed %d %d\n",
            program, pipes [0], pipes [3], extra, pipes [0], pipes [3]);
  }
}

static void my_exec_ad (int * pipes, int num_pipes, int fd)
{
  int i;
  int pid = fork ();
  if (pid == 0) {
    /* ad takes as args the number of pipe pairs, then the actual pipe fds */
    int num_args = 2 /* program + number */ + num_pipes + 1 /* NULL */ ;
    char * * args = malloc_or_fail (num_args * sizeof (char *),
                                    "astart ad args");
    args [0] = "ad";
    args [1] = itoa (num_pipes / 2);
    printf ("calling %s %s ", args [0], args [1]);
    for (i = 0; i < num_pipes / 2; i++) {
      args [2 + i * 2    ] = itoa (pipes [4 * i + 2]);
      args [2 + i * 2 + 1] = itoa (pipes [4 * i + 1]);
      close (pipes [4 * i    ]);
      close (pipes [4 * i + 3]);
      printf ("%s %s ", args [2 + i * 2], args [2 + i * 2 + 1]);
    }
    printf ("\n");
    args [num_args - 1] = NULL;
    execv (args [0], args);
    printf ("error executing %s\n", args [0]);
  } else {  /* parent, close the child pipes */
    print_pid (fd, pid);
    for (i = 0; i < num_pipes / 2; i++) {
      close (pipes [4 * i + 1]);
      close (pipes [4 * i + 2]);
    }
  }
}

static char * pid_file_name ()
{
  int is_root = (geteuid () == 0);
  if (is_root)
    return "/var/run/allnet-pids";
  else
    return "/tmp/allnet-pids";
}

/* returns 0 in case of failure */
static int read_pid (int fd)
{
  int result = -1;
  char buffer [1];
  while (read (fd, buffer, 1) == 1) {
    if ((buffer [0] >= '0') && (buffer [0] <= '9')) {  /* digit */
      if (result == -1)
        result = buffer [0] - '0';
      else
        result = result * 10 + buffer [0] - '0';
    } else if (result == -1) {   /* reading whatever precedes the number */
      /* no need to do anything, just read the next char */
    } else {                     /* done */
      if (result > 1)
        return result;
      printf ("weird result from pid file %d\n", result);
      result = -1;  /* start over */
    }
  }
}

#define AIP_UNIX_SOCKET		"/tmp/allnet-addrs"

static int stop_all ()
{
  char * fname = pid_file_name ();
  int fd = open (fname, O_RDONLY, 0);
  if (fd < 0) {
    printf ("unable to stop allnet daemon, missing pid file %s\n", fname);
    if (geteuid () == 0)
      printf ("if it is running, perhaps it was started as a user process\n");
    else
      printf ("if it is running, perhaps it was started as a root process\n");
    return 1;
  }
  printf ("stopping allnet daemon\n");
  int pid;
  while ((pid = read_pid (fd)) > 0)
    kill (pid, SIGINT);
  close (fd);
  unlink (fname);
  unlink (AIP_UNIX_SOCKET);
}

int main (int argc, char ** argv)
{
  int program_length = strlen (argv [0]);
  char * look_for_stop = argv [0];
  if (program_length > 4)
    look_for_stop = argv [0] + (program_length - 4);
  if (((argc == 1) && (strcmp (look_for_stop, "stop") == 0)) ||
      ((argc == 2) && (strcmp (argv [1],      "stop") == 0)))
    return stop_all ();
  /* two pipes from ad to alocal and back */
  /* two pipes from ad to aip and back */
  /* two pipes from ad to acache and back */
#define NUM_FIXED_PIPES		6
  /* two pipes from ad to each abc and back */
#define NUM_INTERFACE_PIPES	2 
  int num_interfaces = argc - 1;
  int num_pipes = NUM_FIXED_PIPES + NUM_INTERFACE_PIPES * num_interfaces;
  int * pipes = malloc_or_fail (num_pipes * 2 * sizeof (int), "astart pipes");
  init_pipes (pipes, num_pipes);

  int pid_fd = open (pid_file_name (),
                     O_WRONLY | O_TRUNC | O_CREAT | O_CLOEXEC, 0600);
  if (pid_fd < 0) {
    perror ("open");
    printf ("unable to write pids to %s\n", pid_file_name ());
    return 1;
  }

  /* start ad */
  my_exec_ad (pipes, num_pipes, pid_fd);

  /* start all the other programs */
  my_exec2 ("./alocal", pipes, pid_fd);
  my_exec3 ("./aip", pipes + 4, AIP_UNIX_SOCKET, pid_fd);
  my_exec2 ("./acache", pipes + 8, pid_fd);
  int i;
  for (i = 0; i < num_interfaces; i++)
    my_exec3 ("./abc", pipes + 12 + (4 * i), argv [i + 1], pid_fd);
}
