/* config.c: give access to config files */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <dirent.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>

#define ROOT		"~/.allnet"
#define HOME_EXT	"/.allnet"

static void create_dir (char * path)
{
  DIR * d = opendir (path);
  if (d != NULL) {  /* exists, done */
    closedir (d);
    return;
  }

  /* path does not exist, attempt to create it */
  if (mkdir (path, 0777) == 0) /* created */
    return;
  if (errno != ENOENT) { /* some other error, give up */
    perror ("1-mkdir");
    return;
  }

  /* ENOENT: a previous component does not exist, attempt to create it */
  char * last_slash = rindex (path, '/');
  if (last_slash == NULL) /* nothing we can try to create, give up */
    return;
  *last_slash = '\0';
  /* now try to create the parent directory */
  create_dir (path);
  /* and try again to create this directory */
  *last_slash = '/';
  if (mkdir (path, 0777) != 0)
    perror ("mkdir");
}

/* returns the number of characters in the full path name of the given file. */
/* if name is not NULL, also malloc's the string and copies the path into it */
/* returns -1 if the allocation fails */
/* if it allocates the name, also checks to make sure the directory exists,
 * and if not, creates it if possible.  Does not create the file. */
int config_file_name (char * program, char * file, char ** name)
{
  char * root = ROOT;
  int free_root = 0;    /* in case root points to allocated memory */
  int root_length = strlen (ROOT);
  char * allnet_config_env = getenv ("ALLNET_CONFIG");
  if (allnet_config_env != NULL) {
    root = allnet_config_env;
    root_length = strlen (ROOT);
  } else {
    char * home_env = getenv ("HOME");
    if (home_env != NULL) {
      /* printf ("no ALLNET_CONFIG, home is %s\n", home_env); */
      int size = strlen (home_env) + strlen (HOME_EXT) + 1;
      if (name != NULL) {
        char * home = malloc (size);
        if (home == NULL) {
          printf ("unable to allocate %d bytes for home\n", size);
          return -1;
        }
        snprintf (home, size, "%s%s", home_env, HOME_EXT);
        root = home;
        free_root = 1;
        root_length = strlen (home);
      } else {   /* do not allocate, just record the length */
        root_length = size;
      }
    }
  }
  /* printf ("root is %s of length %d (free %d)\n",
          root, root_length, free_root); */

  int total_length = root_length + strlen ("/") + strlen (program)
                   + strlen ("/") + strlen (file) + 1;
  if (name == NULL) {
    if (free_root) free (root);
    return total_length;
  }
  * name = malloc (total_length);
  if (*name == NULL) {
    printf ("unable to allocate %d bytes for config_file_name\n", total_length);
    if (free_root) free (root);
    return -1;
  }
  /* check for the existence of the directory, or create it */
  snprintf (*name, total_length, "%s/%s", root, program);
  create_dir (*name);
  snprintf (*name, total_length, "%s/%s/%s", root, program, file);
  /* printf ("file path is %s\n", *name); */
  if (free_root) free (root);
  return total_length;
}

/* returns a file descriptor, -1 if the file does not exist,
 * or -2 in case of errors */
int open_read_config (char * program, char * file, int print_errors)
{
  char * name;
  int size = config_file_name (program, file, &name);
  if (size < 0)
    return -1;
  int result = open (name, O_RDONLY | O_CREAT, 0600);
  if (result < 0) {
    if (print_errors) {
      if (errno == ENOENT)   /* file not found */
        return -1;
      perror ("open_read_config open");
      printf ("unable to open file %s\n", name);
    }
    free (name);
    return -1;
  }
  free (name);
  return result;
}
