/* listen.c: thread to listen on a port and maintain connected fds */
/*   there is a finite maximum number of fds -- once more are connected, */
/*   old ones are closed */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <pthread.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "packet.h"
#include "util.h"
#include "listen.h"

#ifndef LISTEN_H
/* this structure is declared in the caller and passed to every function.
 * Some of the fields should not be accessed by the caller */
struct listen_info {
  /* accessed by the caller as well as the code in listen.c */
  int num_fds;           /* counter for number of file descriptors in fds */
  int * fds;             /* array of ints for file descriptors */
  pthread_mutex_t mutex; /* mutex for accessing fds */
  pthread_t thread4;     /* listen thread for IPv4 */
  pthread_t thread6;     /* listen thread for IPv6 */
  /* the rest of these fields are for listen.c internal use only */
  int max_num_fds;       /* max number of elements with space in fds */
  char * program_name;   /* e.g. alocal, aip */
  int listen_fd4;        /* fd for listening for new connections on ipv4 */
  int listen_fd6;        /* fd for listening for new connections on ipv6 */
  int port;              /* TCP port number */
  int add_remove_pipe;   /* call add_pipe and remove_pipe */
  /* for testing, make counter a char.  Normally, unsigned int */
  unsigned char counter;  /* cycle counter for least recently used */
  unsigned int * used;   /* array of most recent access times */
};
#endif /* LISTEN_H */

/* returns the fd of the new listen socket, or -1 in case of error */
static int init_listen_socket (char * name, int version, int port, int local)
{
  int isip6 = (version == 6);
  int af = ((isip6) ? AF_INET6 : AF_INET);
  int fd = socket (af, SOCK_STREAM, 0);
  if (fd < 0) {
    perror ("listen socket");
    return -1;
  }
  /* allow us to reuse the port number immediately, rather than wait */
  int option = 1;
  if (setsockopt (fd, SOL_SOCKET, SO_REUSEADDR, &option, sizeof (int)) != 0)
    perror ("setsockopt");

  struct sockaddr_storage address;
  struct sockaddr     * ap  = (struct sockaddr     *) &address;
  struct sockaddr_in  * ap4 = (struct sockaddr_in  *) ap;
  struct sockaddr_in6 * ap6 = (struct sockaddr_in6 *) ap;
  socklen_t addr_size = sizeof (address);

  memset (&address, 0, addr_size);
  if (isip6) {
    ap6->sin6_family = AF_INET6;
    if (local)
      memcpy (&(ap6->sin6_addr), &(in6addr_loopback), sizeof (ap6->sin6_addr));
    else
      memcpy (&(ap6->sin6_addr), &(in6addr_any), sizeof (ap6->sin6_addr));
    ap6->sin6_port = port;
    addr_size = sizeof (struct sockaddr_in6);
  } else {
    ap4->sin_family = AF_INET;
    if (local)
      ap4->sin_addr.s_addr = htonl (INADDR_LOOPBACK);
    else
      ap4->sin_addr.s_addr = htonl (INADDR_ANY);
    ap4->sin_port = port;
    addr_size = sizeof (struct sockaddr_in);
  }
  printf ("binding to ");
  print_sockaddr (ap, addr_size);
  printf ("\n");
  if (bind (fd, ap, addr_size) < 0) {
    if (version == 6) {
      perror ("bind");
      printf ("%s ipv%d unable to bind to %d/%x, probably already running\n",
              name, version, ntohs (port), ntohs (port));
      printf ("bind address is ");
      print_sockaddr (ap, addr_size);
      printf ("\n");
    } else {
      printf ("%s ipv%d unable to bind to %d/%x, probably handled by ipv6\n",
              name, version, ntohs (port), ntohs (port));
    }
    return -1;
  }
  /* specify the maximum queue length */
  if (listen (fd, 5) < 0) {
    perror("listen");
    return -1;
  }
  printf ("%s opened accept socket fd = %d, ip version %d\n", name, fd,
          version);
  return fd;
}

struct real_arg {
  struct listen_info * info;
  int fd;
};

static void * listen_loop (void * arg)
{
  struct real_arg * ra = (struct real_arg *) arg;
  printf ("started listen_loop, fd is %d\n", ra->fd);
  struct listen_info * info = ra->info;

  /* allow the main thread to kill this thread at any time */
  int notinteresting;
  pthread_setcanceltype (PTHREAD_CANCEL_ASYNCHRONOUS, &notinteresting);

  struct sockaddr_storage address;
  struct sockaddr     * ap  = (struct sockaddr     *) &address;
  int addr_size = sizeof (address);

  /* listen for connections, add them to the data structure */
  int connection;
  while ((connection = accept (ra->fd, ap, &addr_size)) >= 0) {
    printf ("%s opened connection socket fd = %d port %d\n",
            info->program_name, connection, ntohs (info->port));
    listen_add_fd (info, connection);

    addr_size = sizeof (address);
  }
  perror ("accept");
  printf ("error calling accept (%d)\n", ra->fd);
  sleep (10);
}

void listen_init_info (struct listen_info * info, int max_fds, char * name,
                       int port, int local_only, int add_remove_pipe)
{
  if (max_fds > 1024) {
    printf ("using 1024 as the maximum number of open fds, %d is too large\n",
            max_fds);
    exit (1);
  }
  if (max_fds <= 0) {
    printf ("invalid %d for max open fds\n", max_fds);
    exit (1);
  }
  info->program_name = name;
  info->port = port;
  info->add_remove_pipe = add_remove_pipe;
  info->num_fds = 0;
  info->max_num_fds = max_fds;
  int bytes = max_fds * (sizeof (int));
  info->fds = malloc_or_fail (bytes, "listen thread fds");
  /* "fds" uses the first part of the allocation, and "used" the second part */
  info->used = malloc_or_fail (bytes, "listen thread used");
  int i;
  for (i = 0; i < max_fds; i++)
    info->fds [i] = info->used [i] = 0;
  info->counter = 0;
  pthread_mutex_init (&(info->mutex), NULL);
  info->listen_fd6 = init_listen_socket (name, 6, port, local_only);
  info->listen_fd4 = init_listen_socket (name, 4, port, local_only);
  if (info->listen_fd6 < 0)
    exit (1);
/*  ipv4 may be handled under ipv6
  if ((info->listen_fd4 < 0) || (info->listen_fd6 < 0))
    exit (1);
*/
  struct real_arg * real_arg6 =
    malloc_or_fail (sizeof (struct real_arg), "ip6 real arg");
  real_arg6->info = info;
  real_arg6->fd = info->listen_fd6;
  if (pthread_create (&(info->thread6), NULL, listen_loop, real_arg6) != 0) {
    perror ("listen6/pthread_create");
    printf ("unable to create listen thread for IP version 6\n");
    exit (1);
  }
  if (info->listen_fd4 >= 0) {
  /* printf ("allocating %ld bytes for 4\n", sizeof (struct real_arg));  */
    struct real_arg * real_arg4 =
      malloc_or_fail (sizeof (struct real_arg), "ip4 real arg");
  /* printf ("allocated %ld bytes\n", sizeof (struct real_arg)); */
    real_arg4->info = info;
    real_arg4->fd = info->listen_fd4;
    if (pthread_create (&(info->thread4), NULL, listen_loop, real_arg4) != 0) {
      perror ("listen4/pthread_create");
      printf ("unable to create listen thread for IP version 4\n");
      exit (1);
    }
  }
}

int listen_record_usage (struct listen_info * info, int fd)
{
  int i;
  if (info->counter + 1 == 0) {  /* wrap around of counter value */
    int decrement = info->counter - (info->counter / 16);
    printf ("wrapping around counter values, decrement %d\n", decrement);
    for (i = 0; i < info->num_fds; i++) {
      printf ("%d: %d", i, info->used [i]);
      if (decrement < info->used [i])
        info->used [i] -= decrement;
      else
        info->used [i] = 0;
      printf (" -> %d\n", info->used [i]);
    }
    info->counter = info->counter - decrement;
  }
  info->counter++;
  for (i = 0; i < info->num_fds; i++)
    if (info->fds [i] == fd)
      info->used [i] = info->counter;
}

/* if some fds are still available, return the next */
/* otherwise, return the index of the oldest FD, after closing it */
/* called with lock held */
static int close_oldest_fd (struct listen_info * info)
{
  if (info->num_fds < info->max_num_fds)
    return info->num_fds++;
  int i;
  int min_index = 0;
  for (i = 1; i < info->num_fds; i++)
    if (info->used [i] < info->used [min_index])
      min_index = i;
  if (info->add_remove_pipe)
    remove_pipe (info->fds [min_index]);
  close (info->fds [min_index]);
  return min_index;
}

void listen_add_fd (struct listen_info * info, int fd)
{
  pthread_mutex_lock (&(info->mutex));
  int index = close_oldest_fd (info);
  info->fds [index] = fd;
  if (info->add_remove_pipe)
    add_pipe (fd);
  pthread_mutex_unlock (&(info->mutex));
}

void listen_remove_fd (struct listen_info * info, int fd)
{
  pthread_mutex_lock (&(info->mutex));
  if (info->add_remove_pipe)
    remove_pipe (fd);
  int i;
  for (i = 0; i < info->num_fds; i++) {
    if (info->fds [i] == fd) {
      info->num_fds--;
      if (i < info->num_fds)
        info->fds [i] = info->fds [info->num_fds];
      break;      /* assume any fd only appears once */
    }
  }
  pthread_mutex_unlock (&(info->mutex));
}

