/* packet.h: definition of AllNet packet headers */

#ifndef PACKET_H
#define PACKET_H

/* allnet keeps two ports: allnet port, for exchanges among peers,
 * and allnet local port, for connection by local clients to the allnet daemon
 */
#define ALLNET_PORT 	     (htons (0xa119))  /* ALLNet */
#define ALLNET_PORT_STRING   ("41241")         /* same */
#define ALLNET_LOCAL_PORT    (htons (0xa11e))  /* ALLnEt */

#define ALLNET_WIFI_PROTOCOL (htons (0xa119))  /* ALLNet */

/* for receiving allnet messages, the receiver must know in advance what the
 * maximum packet size will be.  So allnet defines a maximum size, which
 * includes all headers.  This number is 12345 rounded down to the nearest
 * multiple of 16 */
#define ALLNET_MTU		12336          /* max size for a packet */

/* the allnet basic header has:
 * an 8-bit number b of valid bits in the destination ID
 * an 8-bit number of hops already traversed, 0 when first sent
 * an 8-bit maximum number of hops left
 * an 8-bit next header ID, 0 if there is no next header
 * the 96-bit destination ID
 */

#define DESTINATION_SIZE		12	/* 12 bytes or 96 bits */

struct allnet_header {
/* the first few bits of the packet_type can be used as the protocol
 * version number */
#define ALLNET_TYPE_DATA	1	/* normal encrypted data */
#define ALLNET_TYPE_DATA_ACK	2	/* acknowledge data */
#define ALLNET_TYPE_DATA_REQ	3	/* request data */
#define ALLNET_TYPE_KEY_XCHG	4	/* key exchange */
#define ALLNET_TYPE_CLEAR	5	/* cleartext message */
#define ALLNET_TYPE_MGMT	6	/* AllNet and DHT information */
  unsigned char packet_type;
  unsigned char hops;	   /* times this packet has been recvd, initially 0 */
  unsigned char max_hops;  /* if on receipt hops+1 >= max_hops, do not fwd */
  unsigned char num_destination_bits;
  unsigned char destination [DESTINATION_SIZE];
};

/* 16 bytes (128 bits) */
#define ALLNET_BASIC_HEADER_SIZE	(sizeof (struct allnet_header))


/* the 128-bit packet ID is only used for data, data_ack, and data_req packets,
 * and is only present for these packets.
 * - data packets have in the packet_id field the first PACKET_ID_SIZE
 *   bytes of the sha512 hash of the first PACKET_ID_BYTES bytes of the
 *   decrypted message.  These bytes are usually a random nonce R, and
 *   are placed into the packet_id field of the corresponding ack message.
 * - data request packets have in the packet_id field the destination
 *   address of the packet(s) being requested, preceded by a byte
 *   indicating the number of destination bits.
 */

#define PACKET_ID_SIZE			16	/* 16 bytes or 128 bits */

struct allnet_header_data {
  /* fields from the basic header */
  unsigned char packet_type;     /* DATA, DATA_ACK, or DATA_REQ */
  unsigned char hops;
  unsigned char max_hops;
  unsigned char num_destination_bits;
  unsigned char destination [DESTINATION_SIZE];
  /* additional field */
  unsigned char packet_id [PACKET_ID_SIZE];
};

/* 32 bytes (256 bits) */
#define ALLNET_DATA_HEADER_SIZE		(sizeof (struct allnet_header_data))


#define EXPIRATION_TIME_SIZE	8

/* clear messages may optionally carry: a sender ID, a signature, and
 * an expiration time */
/* the expiration time, if not all 0's, is the number of seconds
 * since midnight, Jan 1, 2000, GMT.  This is encoded in big-endian order */
/* the signature, if present (signature_bytes > 0) is the sender's signature
 * covering all the payload and the header, except the signature and the
 * hops fields (i.e. what is shown in allnet_structure_for_signature) */

struct allnet_header_clear {
  /* fields from the basic header */
  unsigned char packet_type;     /* CLEAR */
  unsigned char hops;
  unsigned char max_hops;
  unsigned char num_destination_bits;
  unsigned char destination [DESTINATION_SIZE];
  /* additional fields, some of them optional */
/* seconds since 00:00 Jan 1, 2000, GMT, or all 0's for no expiration time */
  unsigned char expiration [EXPIRATION_TIME_SIZE];
/* how many bits of the sender are valid.  If > 0, DESTINATION_SIZE bytes
 * of the sender field are sent, even if not all are valid */
  unsigned char sender_id_bits;  /* if not 0, DESTINATION_SIZE bytes sent */
  unsigned char sender [0];      /* 0 or DESTINATION_SIZE */
/* how many bytes the signature has, or 0 for no signature */
  unsigned char signature_bytes;
  unsigned char signature [0];   /* signature_bytes */
};

/* the signature is computed over this structure, with values from the actual
 * header and payload */
struct allnet_structure_for_signature {
  /* fields from the basic header */
  unsigned char num_destination_bits;
  unsigned char destination [DESTINATION_SIZE];
  unsigned char expiration [EXPIRATION_TIME_SIZE];
  unsigned char sender_id_bits;
 /* sender is always DESTINATION_SIZE, set to 0's if sender_id_bits = 0,
  * otherwise exactly as sent, including the invalid bits */
  unsigned char sender [DESTINATION_SIZE];
  unsigned char payload [0];         /* as many bytes as are in the payload */
};

/* If I send a beacon, "destination" indicates what packets I am most
 * interested in receiving, should I wish to give that information */
#define ALLNET_MGMT_BEACON	1	/* sender is ready to receive */
struct allnet_header_mgmt {
  /* fields from the basic header */
  unsigned char packet_type;     /* MGMT */
  unsigned char hops;
  unsigned char max_hops;
  unsigned char num_destination_bits;
  unsigned char destination [DESTINATION_SIZE];
  /* additional fields, some of them optional */
  unsigned char mgmt_type;
};

#define ALLNET_MGMT_HEADER_SIZE	(sizeof (struct allnet_header_mgmt))

#endif /* PACKET_H */
