/* pipemsg.c: transmit messages over pipes */

/* all but the receive functions return 1 in case of success and 0
 * in case of failure.
 * if 0 is returned, it means the pipe is no longer valid.
 * the receive functions return the number of bytes received, which
 * for receive_pipe_message_buffer may be greater than mlen.
 * in case of failure, the receive functions return -1 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <sys/select.h>

#include "packet.h"
#include "priority.h"
#include "pipemsg.h"

#define MAGIC_STRING	"MAGICPIPE"

/* MAGIC_SIZE does not include null character at the end */
#define MAGIC_SIZE	(sizeof (MAGIC_STRING) - 1)
#define PRIORITY_SIZE	4
#define LENGTH_SIZE	4
#define HEADER_SIZE	(MAGIC_SIZE + PRIORITY_SIZE + LENGTH_SIZE)

static int num_pipes = 0;

struct pipe_info {
  int pipe_fd;	   /* file descriptor for input */
  int in_header;
  char header [HEADER_SIZE];
  char * buffer;   /* may be null */
  int filled;      /* how many bytes does the buffer or header already have */
  int bsize;       /* how many bytes do we receive before we are done? */
};

static struct pipe_info * buffers = NULL;
static int * just_pipes = NULL;

static int do_not_print = 0;
static void print_pipes (const char * desc, int pipe)
{
  if (do_not_print)
    return;
  printf ("%s pipe %d, total %d:\n", desc, pipe, num_pipes);
  int i;
  for (i = 0; i < num_pipes; i++)
    printf ("  [%d]: pipe %d, %shdr, h %p b %p, filled %d bsize %d\n", i,
            buffers [i].pipe_fd, ((buffers [i].in_header) ? "" : "not "),
            buffers [i].header, buffers [i].buffer,
            buffers [i].filled, buffers [i].bsize);
  printf ("pipes: ");
  for (i = 0; i < num_pipes; i++)
    printf (" %d", just_pipes [i]);
  printf ("\n");
}

void add_pipe (int pipe)
{
  int i;
  for (i = 0; i < num_pipes; i++) {
    if (buffers [i].pipe_fd == pipe) {
      printf ("adding pipe %d already in data structure [%d]\n", pipe, i);
      return;
    }
  }
  int total = num_pipes + 1;
  size_t size = total * sizeof (struct pipe_info);
  struct pipe_info * new_buffer = (struct pipe_info *) malloc (size);
  if (new_buffer == NULL)
    printf ("unable to allocate %zd bytes for %d slots\n", size, total);
  int * new_pipes = (int *) malloc (total * sizeof (int));
  if (new_pipes == NULL)
    printf ("unable to allocate %zd bytes for %d pipes\n",
            total * sizeof (int), total);
  if ((new_buffer == NULL) || (new_pipes == NULL))
    return;
  for (i = 0; i < num_pipes; i++) {
    new_buffer [i] = buffers [i];
    if (new_buffer [i].in_header)  /* update the pointer */
      new_buffer [i].buffer = new_buffer [i].header;
    new_pipes [i] = just_pipes [i];
  }
  new_buffer [num_pipes].pipe_fd = pipe;
  new_buffer [num_pipes].in_header = 1;/* always start by reading the header */
  new_buffer [num_pipes].buffer = new_buffer [num_pipes].header;
  new_buffer [num_pipes].bsize = HEADER_SIZE;
  new_buffer [num_pipes].filled = 0;
  new_pipes [num_pipes] = pipe;

  if (buffers != NULL)
    free (buffers);
  if (just_pipes != NULL)
    free (just_pipes);
  buffers = new_buffer;
  just_pipes = new_pipes;
  num_pipes = total;
  print_pipes ("added", pipe);
}

void add_pipes (int * pipes, int npipes)
{
  do_not_print = 1;
  int i;
  for (i = 0; i < npipes; i++)
    add_pipe (pipes [i]);
  do_not_print = 0;
  print_pipes ("add_pipes", npipes);
}

void remove_pipe (int pipe)
{
  int i;
  for (i = 0; i < num_pipes; i++) {
    if (buffers [i].pipe_fd == pipe) {
      printf ("removing pipe %d from data structure [%d]\n", pipe, i);
      if ((! buffers [i].in_header) && (buffers [i].buffer != NULL))
        free (buffers [i].buffer);
      if (i + 1 < num_pipes) {
        buffers [i] = buffers [num_pipes - 1];
        just_pipes [i] = just_pipes [num_pipes - 1];
        if (buffers [i].in_header)  /* update the pointer */
          buffers [i].buffer = buffers [i].header;
      }
      num_pipes--;
      print_pipes ("removed", pipe);
      return;
    }
  }
}

static int send_pipe_message_orig (int pipe, char * message, int mlen,
                                   int priority)

{
  char header [HEADER_SIZE];
  memcpy (header, MAGIC_STRING, MAGIC_SIZE);

  header [MAGIC_SIZE + 0] = (priority >> 24) & 0xff;
  header [MAGIC_SIZE + 1] = (priority >> 16) & 0xff;
  header [MAGIC_SIZE + 2] = (priority >>  8) & 0xff;
  header [MAGIC_SIZE + 3] =  priority        & 0xff;

  header [MAGIC_SIZE + 4] = (mlen >> 24) & 0xff;
  header [MAGIC_SIZE + 5] = (mlen >> 16) & 0xff;
  header [MAGIC_SIZE + 6] = (mlen >>  8) & 0xff;
  header [MAGIC_SIZE + 7] =  mlen        & 0xff;

  int w = write (pipe, header, HEADER_SIZE); 
  if (w != HEADER_SIZE) {
    perror ("send_pipe_msg(1) write");
    printf ("(1) pipe number %d, result %d\n", pipe, w);
    return 0;
  }

  w = write (pipe, message, mlen); 
  if (w != mlen) {
    perror ("send_pipe_msg(2) write");
    printf ("(2) pipe number %d\n", pipe);
    return 0;
  }
  return 1;
}

int send_pipe_message (int pipe, char * message, int mlen, int priority)
{
  /* avoid SIGPIPE signals when writing to a closed pipe */
  struct sigaction sa;
  sa.sa_handler = SIG_IGN;
  sa.sa_flags = 0;
  sa.sa_mask;
  sigemptyset (&(sa.sa_mask));
  struct sigaction old_sa;
  sigaction (SIGPIPE, &sa, &old_sa);

  int result = send_pipe_message_orig (pipe, message, mlen, priority);

  sigaction (SIGPIPE, &old_sa, NULL);
  return result;
}

/* same as send_pipe_message, but frees the memory referred to by message */
int send_pipe_message_free (int pipe, char * message, int mlen, int priority)
{
  int result = send_pipe_message (pipe, message, mlen, priority);
  printf ("send_pipe_message_free freeing %p\n", message);
  free (message);
  printf ("send_pipe_message_free freed %p\n", message);
  return result;
}

/* receives the message into the buffer.  If the message doesn't fit */
/* in the buffer, the remainder is discarded. */
int receive_pipe_message_buffer (int pipe, char * message, int mlen,
                                 int * priority)
{
  char * new_buf;
  int result = receive_pipe_message (pipe, &new_buf, priority);
  if (result < 0)
    return result;
  if (result <= mlen)
    memcpy (message, new_buf, result);
  else
    memcpy (message, new_buf, mlen);
  free (new_buf);
  return result;
}

/* returns the number of characters received, or -1 in case of error */
/* if may_block is false, only returns -1 or blen. */
/* if may_block is true, may return any value between -1 and blen */
static int receive_bytes (int pipe, char * buffer, int blen, int may_block)
{
  int recvd = 0;
  while (recvd < blen) {
    if (! may_block) {
      if (next_available (&pipe, 1, PIPE_MESSAGE_NO_WAIT) < 0)
        return recvd;   /* timed out */
    }
    /* if we did not call next_available, the call to read may block */
/*  printf ("read (%d, %p, %d) => ", pipe, buffer + recvd, blen - recvd); */
    int new_recvd = read (pipe, buffer + recvd, blen - recvd);
/*  printf ("%d\n", new_recvd); */
    if (new_recvd <= 0) {
      if (new_recvd < 0)
        perror ("pipemsg.c receive_bytes read");
      printf ("received %d bytes on pipe %d, expected %d/%d\n",
              new_recvd, pipe, blen - recvd, blen);
      if ((new_recvd == 0) && (recvd > 0))
        return recvd; /* return this data for now, and -1 next time */
      /* not received anything yet -- error or closed pipe */
      return -1;  /* error */
    }
    recvd += new_recvd;
  }
  return blen;
}

/* returns -1 if the header is not valid, and the data size otherwise */
/* note: does not work for buffers of size 2^32 - 1 -- but any buffer
 * larger than 2^31 - 1 is likely to cause problems due to signed ints anyway */
static int parse_header (char * header, int pipe, int * priority)
{
  static int printed = 0;
  if (memcmp (header, MAGIC_STRING, MAGIC_SIZE) != 0) {
    if (printed == 0) {
      printf ("error: unsynchronized pipe %d\n", pipe);
      /* print_buffer (header, HEADER_SIZE, "header", 20, 1); */
    }
    printed++;
    return -1;
  }
  if (printed != 0)
    printf ("  (%dx)\n", printed);
  printed = 0;
  int received_priority = (header [MAGIC_SIZE    ] & 0xff) << 24 |
                          (header [MAGIC_SIZE + 1] & 0xff) << 16 |
                          (header [MAGIC_SIZE + 2] & 0xff) <<  8 |
                          (header [MAGIC_SIZE + 3] & 0xff);
  int received_length   = (header [MAGIC_SIZE + 4] & 0xff) << 24 |
                          (header [MAGIC_SIZE + 5] & 0xff) << 16 |
                          (header [MAGIC_SIZE + 6] & 0xff) <<  8 |
                          (header [MAGIC_SIZE + 7] & 0xff);
  *priority = received_priority;
  return received_length;
}

/* shift the header left by one position, to make room for one more char */
static void shift_header (char * header)
{
  int i;
  for (i = 0; i + 1 < HEADER_SIZE; i++)
    header [i] = header [i + 1];
}

/* receives the message into a buffer it allocates for the purpose. */
/* the caller is responsible for freeing the buffer. */
int receive_pipe_message (int pipe, char ** message, int * priority)
{
  char header [HEADER_SIZE];

  int wanted = HEADER_SIZE;
  int received_len = 0;
  while (1) {
    /* read the header */
    int r = receive_bytes (pipe, header + (HEADER_SIZE - wanted), wanted, 1);
    if (r < 0)
      return -1;
    if (r == 0)
      return 0;
    /* first part of header should be MAGIC_STRING.  If not, it is an error,
       report it and keep looking */
    received_len = parse_header (header, pipe, priority);
    if (received_len != -1)   /* successfully received the header */
      break;
    shift_header (header);    /* no match, shift the header and try again */
    wanted = 1;
  }

  /* allocate the result buffer */
  char * buffer = malloc (received_len);
  if (buffer == NULL) {
    printf ("unable to allocate %d bytes for receive_pipe_message buffer\n",
            received_len);
    return -1;
  }
  /* printf ("receive_pipe_message allocated buffer %p\n", buffer); */

  if (receive_bytes (pipe, buffer, received_len, 1) < 0) {
    free (buffer);
    return -1;
  }

  /* return the buffer and length to the caller */
  *message = buffer;
  return received_len;
}

/* the next function is like receive_pipe_message,
 * but may return 0 if no message is immediately ready to return.
 * it uses select/poll with a timeout of 0 to check for data availability.
 * because external buffer management is more complicated when receiving
 * partial messages, there is no equivalent to receive_pipe_message_buffer. */
int receive_pipe_message_poll (int pipe, char ** message, int * priority)
{
  char * null = NULL;
  int pipe_index = 0;
  while (1) {
    if (pipe_index >= num_pipes) {
      print_pipes ("not found", pipe);
      *null = 'a';
      return -1;
    }
    if (pipe == buffers [pipe_index].pipe_fd)
      break;
    pipe_index++;
  }

  int offset = buffers [pipe_index].filled;
/*
  printf ("calling receive_bytes (%d, %p, %d, 1)\n",
          pipe, buffers [pipe_index].buffer + offset,
          buffers [pipe_index].bsize - offset, 0);
*/
  int read = receive_bytes (pipe, buffers [pipe_index].buffer + offset,
                            buffers [pipe_index].bsize - offset, 0);
/*
  printf ("receive_pipe_message_poll got %d bytes on pipe %d\n", read, pipe);
  print_buffer (buffers [pipe_index].buffer + offset, read, "b", 20, 1);
*/
  if (read <= 0) {
  /* printf ("receive_pipe_message_poll error %d on pipe %d\n", read, pipe); */
    return read;
  }
  offset += read;
  if (offset == buffers [pipe_index].bsize) { /* received all we were looking for */
    if (buffers [pipe_index].in_header) {
      int received_len =
        parse_header (buffers [pipe_index].header, pipe, priority);
      if (received_len != -1) {  /* successfully read a header */
/*      printf ("preparing to receive %d bytes on pipe %d\n", received_len,
                pipe); */
        buffers [pipe_index].buffer = malloc (received_len);
        if (buffers [pipe_index].buffer == NULL) {
          printf ("failed allocating %d for receive_pipe_message_poll\n",
                  received_len);
          return -1;
        }
        buffers [pipe_index].bsize = received_len;
        buffers [pipe_index].in_header = 0;
        buffers [pipe_index].filled = 0;
      } else {  /* no magic_string at start of header */
        /* shift header and try again */
        shift_header (buffers [pipe_index].header);
        buffers [pipe_index].filled--;
      }
      /* we stopped because the header was filled.  Try again, for either
         the next attempt to match, or the actual data */
      return receive_pipe_message_poll (pipe, message, priority);
    } else {    /* not reading header and buffer is full, so we are done. */
      *message = buffers [pipe_index].buffer;
      int result = buffers [pipe_index].bsize;
      buffers [pipe_index].in_header = 1;  /* next time read the header again */
      buffers [pipe_index].buffer = buffers [pipe_index].header;
      buffers [pipe_index].bsize = HEADER_SIZE;
      buffers [pipe_index].filled = 0;
      return result;
    }
  } /* else  buffer is not filled, just return to caller */
  return 0;
}

static void add_us (struct timeval * t, long long us)
{
  us += t->tv_usec;
  t->tv_usec = us % 1000000LL;
  t->tv_sec += us / 1000000LL;
}

static int tv_compare (struct timeval * t1, struct timeval * t2)
{
/*
  printf ("comparing %ld.%06ld to %ld.%06ld\n",
          t1->tv_sec, t1->tv_usec, t2->tv_sec, t2->tv_usec);
*/
  if (t1->tv_sec < t2->tv_sec)
    return -1;
  if (t1->tv_sec > t2->tv_sec)
    return 1;
  if (t1->tv_usec < t2->tv_usec)
    return -1;
  if (t1->tv_usec > t2->tv_usec)
    return 1;
  return 0;
}

/* call receive_pipe_message_poll on a number of pipes, returning the
 * size and message for the first one received (if any).  The pipe from which
 * the message is received is returned in *from_pipe.  */
/* if udp_fd is not -1, also listens to the UDP fd, returning the
 * corresponding message (assumed less than or equal to ALLNET_MTU) */
/* timeout is in ms, or one of the values above to not wait or wait forever */
/* returns 0 in case of timeout */
/* returns -1 and sets *from_pipe in case of error on a pipe. */
int receive_pipe_message_any (int timeout, char ** message, int udp_fd,
                              int * from_pipe, int * priority)
{
  *from_pipe = -1;  /* set it to an invalid value in case we return error */
  struct timeval tv;
  struct timeval * tvp = &tv;
  struct timeval now, finish;
  gettimeofday (&now, NULL);
  finish = now;
  if (timeout != PIPE_MESSAGE_WAIT_FOREVER)
    add_us (&finish, timeout * 1000);

  while ((timeout == PIPE_MESSAGE_WAIT_FOREVER) ||
         (tv_compare (&now, &finish) <= 0)) {
#ifdef DEBUG_PRINT
    int delta = (finish.tv_sec - now.tv_sec) * 1000000 +
                (finish.tv_usec - now.tv_usec);
    printf ("now is %d.%06d, finish is %d.%06d (delta %d.%06d)\n",
            (int) now.tv_sec, (int) now.tv_usec,
            (int) finish.tv_sec, (int) finish.tv_usec,
            delta / 1000000, delta % 1000000);
    printf ("  tvp is %p", tvp);
    if (tvp != NULL)
      printf (" = %d.%06d", (int) tvp->tv_sec, (int) tvp->tv_usec);
    printf ("\n");
#endif /* DEBUG_PRINT */

    int pipe;
    if (udp_fd < 0)
      pipe = next_available (just_pipes, num_pipes, timeout);
    else
      pipe = next_available2 (just_pipes, num_pipes, udp_fd, timeout);
#ifdef DEBUG_PRINT
    printf ("now is %d, finish is %d (delta %d), result of next_avl is %d\n",
            (int) now.tv_sec, (int) finish.tv_sec,
            (int) (finish.tv_sec - now.tv_sec), pipe);
#endif /* DEBUG_PRINT */
    if (pipe >= 0) { /* can read pipe */
      int r;
      if (pipe != udp_fd) {  /* it is a pipe, not a UDP fd */
#ifdef DEBUG_PRINT
        printf ("calling receive_pipe_message_poll (%d, %p, %p)\n",
                pipe, message, priority);
#endif /* DEBUG_PRINT */
        r = receive_pipe_message_poll (pipe, message, priority);
#ifdef DEBUG_PRINT
        if (r > 0)
          printf ("received %d bytes from pipe %d\n", r, pipe);
        else
          printf ("received %d from pipe %d\n", r, pipe);
#endif /* DEBUG_PRINT */
        if ((r != 0)) {
          *from_pipe = pipe;
          return r;
        }   /* else did not read a complete buffer, continue the loop */
            /* until timeout */
      } else {           /* UDP socket, return availability */
#ifdef DEBUG_PRINT
        printf ("udp socket %d, returning 0\n", pipe);
#endif /* DEBUG_PRINT */
        *from_pipe = pipe;
        return 0;
      }
    }
    gettimeofday (&now, NULL);
  }
  return 0;    /* timed out */
}

/* returns the first available file descriptor, or -1 in case of timeout */
/* timeout is in milliseconds, or one of the values above */
int next_available (int * pipes, int npipes, int timeout)
{
  struct timeval tv;
  struct timeval * tvp = &tv;
  if (timeout == PIPE_MESSAGE_WAIT_FOREVER) {
    /* wait forever, so don't give a timeout */
    tvp = NULL;
  } else {
   /* if timeout is PIPE_MESSAGE_NO_WAIT, timeout is 0, which
    * works just right in this computation */
    tvp->tv_sec = timeout / 1000;
    tvp->tv_usec = (timeout % 1000) * 1000;
  }

#ifdef DEBUG_PRINT
  printf ("  tvp is %p", tvp);
  if (tvp != NULL)
    printf (" = %d.%06d", (int) tvp->tv_sec, (int) tvp->tv_usec);
  printf ("\n");
#endif /* DEBUG_PRINT */

  /* set up the readfd bitset */
  int max_pipe = 0;
  fd_set receiving;
  FD_ZERO (&receiving);
  int i;
  for (i = 0; i < npipes; i++) {
    FD_SET (pipes [i], &receiving);
    if (pipes [i] > max_pipe)
      max_pipe = pipes [i];
#ifdef DEBUG_PRINT
    printf ("selecting pipe %d [%d]\n", pipes [i], i);
#endif /* DEBUG_PRINT */
  }

  /* call select */
#ifdef DEBUG_PRINT
  printf ("calling select (%d)\n", max_pipe + 1);
#endif /* DEBUG_PRINT */
  int s = select (max_pipe + 1, &receiving, NULL, NULL, tvp);
#ifdef DEBUG_PRINT
  printf ("result of select is %d\n", s);
#endif /* DEBUG_PRINT */
  if (s < 0) {
    perror ("next_available/select");
    print_pipes ("current", max_pipe);
    for (i = 0; i < npipes; i++)
      printf ("pipes [%d] is %d\n", i, pipes [i]);
    printf ("some error in select, aborting\n");
    exit (1);
  }
  if (s == 0)
    return -1;
  if (s > 0) {
#ifdef DEBUG_PRINT
    for (i = 0; i < npipes; i++) {
      if (FD_ISSET (pipes [i], &receiving))
        printf (" +%d", buffers [i].pipe_fd);
      else
        printf (" -%d", buffers [i].pipe_fd);
    }
    printf ("\n");
#endif /* DEBUG_PRINT */
/* this loop favors pipes earlier in the array.  Not a big deal, I think */
    for (i = 0; i < npipes; i++) {
#ifdef DEBUG_PRINT
      if (FD_ISSET (pipes [i], &receiving))  /* has some data */
        printf ("next_available returning %d\n", pipes [i]);
#endif /* DEBUG_PRINT */
      if (FD_ISSET (pipes [i], &receiving))  /* has some data */
        return pipes [i];
    }
  }
  printf ("next_available: strange, s is %d but no pipes found\n", s);
  return 0;
}

/* returns the first available file descriptor, or -1 in case of timeout */
/* timeout is in milliseconds, or one of the values above */
int next_available2 (int * fds, int nfds, int udp_fd, int timeout)
{
  static int * both = NULL;
  static int nboth = 0;

  if (nfds + 1 > nboth) {
    if (both != NULL) free (both);
    nboth = nfds + 1;
    both = malloc (sizeof (int) * nboth);
    if (both == NULL) {
      perror ("pipemsg.c/available2/malloc");
      exit (1);
    }
  }
  int index;
  for (index = 0; index < nfds; index++)
    both [index] = fds [index];
  both [nfds] = udp_fd;

  return next_available (both, nfds + 1, timeout);
}

