/* pipemsg.h: transmit messages over pipes */

#ifndef PIPEMSG_H
#define PIPEMSG_H

/* the send functions return 1 in case of success and 0 in case of failure.
 * if 0 is returned, it means the pipe is no longer valid.
 * the receive functions return the number of bytes received, 0
 * in case of timeout, and -1 in case of error, including the pipe
 * no longer being usable.
 * receive_pipe_message_buffer may return more than mlen, but only fills in
 * up to mlen bytes of the buffer (so use receive_pipe_message instead!). */

/* keeps track of which pipes are needed for
 * receive_pipe_message_poll and receive_pipe_message_any */
/* this code does not receive on do_not_receive pipes, only returns them */
extern void add_pipe (int pipe);
extern void add_pipes (int * pipes, int npipes);
extern void remove_pipe (int pipe);

extern int send_pipe_message (int pipe, char * message, int mlen, int priority);

/* same as send_pipe_message, but frees the memory referred to by message */
extern int send_pipe_message_free (int pipe, char * message, int mlen,
                                   int priority);

#define PIPE_MESSAGE_WAIT_FOREVER	-1
#define PIPE_MESSAGE_NO_WAIT		0
/* returns the first available file descriptor, or -1 in case of timeout */
/* timeout is in milliseconds, or one of the values above */
extern int next_available (int * pipes, int npipes, int timeout);

/* receives the message into the buffer.  If the message doesn't fit */
/* in the buffer, the remainder is discarded. */
extern int receive_pipe_message_buffer (int pipe, char * message, int mlen,
                                        int * priority);

/* receives the message into a buffer it allocates for the purpose. */
/* the caller is responsible for freeing the buffer. */
extern int receive_pipe_message (int pipe, char ** message, int * priority);

/* the next function is like receive_pipe_message,
 * but may return 0 if no message is immediately ready to return.
 * it uses select/poll with a timeout of 0 to check for data availability.
 * because external buffer management is more complicated when receiving
 * partial messages, there is no equivalent to receive_pipe_message_buffer. */
/* also, pipe_index is the index into the pipe array, not the actual pipe */
extern int receive_pipe_message_poll (int pipe, char ** message,
                                      int * priority);

/* call receive_pipe_message_poll on all known pipes, returning the
 * size and message for the first one received (if any).  The pipe from which
 * the message is received is returned in *from_pipe.  */
/* timeout is in ms, or one of the values above to not wait or wait forever */
/* if udp_fd is not -1, also listens to the UDP fd, returning the
 * corresponding message (assumed less than or equal to ALLNET_MTU) */
/* timeout is in ms, or one of the values above to not wait or wait forever */
/* returns 0 in case of timeout */
/* returns -1 and sets *from_pipe in case of error on a pipe. */
extern int receive_pipe_message_any (int timeout, char ** message, int udp_fd,
                                     int * from_pipe, int * priority);

#endif /* PIPEMSG_H */
