/* record.c: keep track of recently received packets */

/* the way we keep track is to use two hash tables.  Each hash table
 * is indexed by two bytes of the hash of the packet, and records
 * the entire hash.  New hashes overwrite old hashes, but presumably
 * in different locations in the two hash tables, so in at least one
 * of the hash tables, the hash is likely to remain for a while.
 * When looking up, a match in either hash table means we have seen
 * the packet recently.
 */

#include <stdio.h>

#define ENTRIES_PER_TABLE		1024

static int hash1 [ENTRIES_PER_TABLE];
static int hash2 [ENTRIES_PER_TABLE];

/* data must have at least ((bits + 7) / 8) bytes */
int my_hash_fn (char * data, int bits)
{
  /* just xor all the bytes in the data, shifting as we go along.
   * if there are any odd (< 32) bits at the end, they are used
   * to initialize the result
   */
  int * idata = ((int *) data);
  int isize = bits / 32;
  int i;
  int result = 0;
  if (isize * 32 < bits) {
    result = data [isize * 4];
    if (isize * 32 + 8 < bits)
      result |= ((data [isize * 4 + 1]) << 8);
    if (isize * 32 + 16 < bits)
      result |= ((data [isize * 4 + 2]) << 16);
    if (isize * 32 + 24 < bits)
      result |= ((data [isize * 4 + 3]) << 24);
  }
  result += bits;  /* include the size of the packet */
  for (i = 0; i < isize; i++) {
    result = (result << 1) | ((result >> 31) & 1);
    result = result ^ idata [i];
  }
  if (result < 0)
    result = - result;
  return result;
}

/* return 0 if this is a new packet, and 1 if it has been seen before */
int record_packet (char * data, int dsize)
{
  static int initialized = 0;
  if (! initialized) {
    int i;
    for (i = 0; i < ENTRIES_PER_TABLE; i++) {
      hash1 [i] = 0;
      hash2 [i] = 0;
    }
    initialized = 1;
  }
  int hash = my_hash_fn (data, dsize * 8);
  int left_hash  = ((hash >> 16) & 0xff00) | ((hash >> 8) & 0xff);
  int right_hash = ((hash >>  8) & 0xff00) | ( hash       & 0xff);
/*
  printf ("hash = %x, left %d %x, right = %d %x\n", hash,
          left_hash % ENTRIES_PER_TABLE,
          hash1 [left_hash  % ENTRIES_PER_TABLE],
          right_hash % ENTRIES_PER_TABLE,
          hash2 [right_hash  % ENTRIES_PER_TABLE]);
 */
  if ((hash1 [left_hash  % ENTRIES_PER_TABLE] == hash) ||
      (hash2 [right_hash % ENTRIES_PER_TABLE] == hash))
    return 1;  /* found */
  /* store into both hash tables */
  hash1 [left_hash  % ENTRIES_PER_TABLE] = hash;
  hash2 [right_hash % ENTRIES_PER_TABLE] = hash;
  return 0;  /* not found */
}
