/* basic_ping.c:
 * send a message to the allnet local port, see if we get it back
 *
 * if an IP address (and optionally a port number) is given as a parameter,
 * send the address to allnet as a mapping for the destination address
 * we will use.  That basically sends the message to the recipient address,
 * which, if running, will forward to any attached clients.  If you
 * know enough to understand this, it allows you to send a very basic ping.
 *
 * if you don't understand it, please wait for a better client.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <netinet/in.h>

#include "../packet.h"
#include "../pipemsg.h"
#include "../priority.h"
#include "../util.h"

/* originally copied from tsocial.c */
static void init_array (char * data, int first, int count)
{
  int i;
  for (i = 0; i < count; i++)
    data [i] = (first + i) & 0xff;
}

static void init_packet (char * p, int psize, int type,
                         int hops, int max_hops, int nbits, int first)
{
  p [0] = type;
  p [1] = hops;
  p [2] = max_hops;
  p [3] = nbits;
  init_array (p + 4, first, psize - 4);
  print_buffer (p, psize, "packet", 18, 1);
}

static void receive_all_messages (char * dest, int sent_size)
{
  int loops = 0;
  while (1) {
    char * message;
    int pipe, pri;
    int r = receive_pipe_message_any (10000, &message, -1, &pipe, &pri);
    if (r <= 0) {
      if (loops > 0) printf ("no more messages\n");
      else           printf ("no messages\n");
      break;
    }
    if ((r >= sent_size - 1) &&     /* this is our message or the reply */
        (memcmp (dest, message + 4, 4) == 0)) {
      struct timeval finish;
      gettimeofday (&finish, NULL);
      struct timeval * start = (struct timeval *) (message + 100);
      int delta_s = finish.tv_sec - start->tv_sec;
      int delta_us = finish.tv_usec - start->tv_usec;
      while (delta_us < 0) {
        delta_us += 1000000;
        delta_s--;
      }
      if (r == sent_size)
        printf ("got local reply, time %d.%06d, %d bytes\n",
                delta_s, delta_us, r);
      else
        printf ("got remote reply, time %d.%06d, %d bytes\n",
                delta_s, delta_us, r);
    } else if ((loops < 3) && (r == sent_size)) {   /* from peer, reply */
      send_pipe_message (pipe, message, r - 1, ONE_HALF);
      printf ("sent reply, %d bytes on pipe %d\n", r - 1, pipe);
    } else {
      printf ("received %d bytes on pipe %d", r, pipe);
      printf (", priority ");
      print_fraction (pri, NULL);
      printf ("\n");
      print_buffer (message, r, NULL, 20, 1);
    }
    free (message);
    loops++;
  }
}

static int new_local_socket (int ipv6)
{
  int fd = socket (ipv6 ? AF_INET6 : AF_INET, SOCK_STREAM, 0);
  if (fd < 0) {
    perror ("socket");
    exit (1);
  }
  struct sockaddr_storage address;
  struct sockaddr     * ap  = (struct sockaddr     *) &address;
  struct sockaddr_in  * ap4 = (struct sockaddr_in  *) ap;
  struct sockaddr_in6 * ap6 = (struct sockaddr_in6 *) ap;
  socklen_t addr_size = sizeof (address);

  if (ipv6) {
    ap6->sin6_family = PF_INET6;
    memcpy (&(ap6->sin6_addr), &(in6addr_loopback), sizeof (ap6->sin6_addr));
    ap6->sin6_port = ALLNET_LOCAL_PORT;
  } else {
    ap4->sin_family = AF_INET;
    ap4->sin_addr.s_addr = htonl (INADDR_LOOPBACK);
    ap4->sin_port = ALLNET_LOCAL_PORT;
  }
  if (connect (fd, ap, addr_size) < 0) {
    if (errno == ECONNREFUSED)
      printf ("nobody listening on port %d (allnet daemon is not running)\n",
              ntohs (ALLNET_LOCAL_PORT));
    else
      perror ("connect");
    exit (1);
  }
  return fd;
}

/* copied from aip.c -- should be in a header somewhere */
struct addr_info {
  unsigned char destination [DESTINATION_SIZE];
  unsigned char nbits;    /* how many bits of the destination are given */
  unsigned char ip_version;   /* 4 or 6 */
  unsigned short port;
  struct in6_addr ip;         /* for ip4, only 4 bytes are filled/sent */
};

static void send_mapping (int argc, char ** argv, int init, int nbits)
{
  if (argc > 1) {
    int addr_socket = socket (AF_UNIX, SOCK_DGRAM, 0);
    if (addr_socket < 0) {
      perror ("af_unix socket");
      exit (1);    /* not sure what this error would mean */
    }
    struct sockaddr_un sun;
    sun.sun_family = AF_UNIX;
    snprintf (sun.sun_path, sizeof (sun.sun_path), "%s", "/tmp/allnet-addrs");
    if (connect (addr_socket, (struct sockaddr *) (&sun), sizeof (sun)) < 0) {
      perror ("unix socket connect");
      printf ("allnet addrs socket is missing, allnet may not be running\n");
      exit (1);
    }

    struct addr_info ai;
    init_array (ai.destination, init, DESTINATION_SIZE);
    ai.nbits = nbits;
    if (inet_pton (AF_INET6, argv [1], &(ai.ip)) > 0)
      ai.ip_version = 6;
    else if (inet_pton (AF_INET, argv [1], &(ai.ip)) > 0)
      ai.ip_version = 4;
    else {
      printf ("error: unable to convert '%s' to an IP address\n", argv [1]);
    }
    ai.port = ALLNET_PORT;
    if (argc > 2)
      ai.port = htons (atoi (argv [2]));
    if (ai.ip_version == 6)
      send_pipe_message (addr_socket, (char *) (&ai), sizeof (ai), ONE_HALF);
    else
      send_pipe_message (addr_socket, (char *) (&ai), sizeof (ai) - 12,
                         ONE_HALF);
    close (addr_socket);
    /* sleep (2); */
  }
}

int main (int argc, char ** argv)
{
  int sock = new_local_socket (1);
  add_pipe (sock);

  int now = time (NULL);
  send_mapping (argc, argv, now, 48);

  char packet [200];
  init_packet (packet, sizeof (packet), ALLNET_TYPE_CLEAR, 0, 1, 48, now);
  printf ("sending clear packet, 0 -> 1 hops, 1/2 priority\n");
  struct timeval start;
  gettimeofday (&start, NULL);
  memcpy (packet + 100, (char *) (&start), sizeof (start));
  send_pipe_message (sock, packet, sizeof (packet), ONE_HALF);
  receive_all_messages (packet + 4, sizeof (packet));
  return 0;
}
