/* tprirectrk.c: test priority.c, record.c, and track.c */

#include <stdio.h>
#include <string.h>

#include "../packet.h"
#include "../priority.h"
#include "../record.h"
#include "../track.h"

static void priority_test (int computed, int expected, char * desc)
{
  if (computed == expected) {
    printf ("%s priority %x (", desc, computed);
    print_fraction (computed, NULL);
    printf (") matches expected %x (", expected);
    print_fraction (expected, NULL);
    printf (")\n");
  } else {
    printf ("%s priority %x (", desc, computed);
    print_fraction (computed, NULL);
    printf (") was expected to be %x (", expected);
    print_fraction (expected, NULL);
    int diff = computed - expected;
    printf ("), difference is %x (", diff);
    print_fraction (diff, NULL);
    printf (")\n");
  }
}

static void observe_priority (int is_local, int size, int dest_bits,
                              int hops_already, int hops_left,
                              int social_distance, int rate_fraction,
                              char * desc)
{
  int p = compute_priority (is_local, size, dest_bits, hops_already, hops_left,
                            social_distance, rate_fraction);
  printf ("%s priority %x (", desc, p);
  print_fraction (p, NULL);
  printf (")\n");
}

static void record_test (char * data, int dlen, int expected)
{
  if (expected) {
    if (! record_packet (data, dlen)) {
      printf ("error: previously seen packet not found\n");
    } else {
      printf ("record: previously seen packet correctly detected\n");
    }
  } else {
    if (record_packet (data, dlen)) {
      printf ("error: new packet was found\n");
    } else {
      printf ("record: new packet accepted\n");
    }
  }
}

/* copied from tsocial.c */
static void init_dest (char * dest, int first, int count)
{
  int i;
  for (i = 0; i < count; i++)
    dest [i] = (first + i) & 0xff;
  printf ("%d bytes:", count);
  for (i = 0; i < count && i < 20; i++)
    printf (" %02x", dest [i] & 0xff);
  if (i < count)
    printf (" ...");
  printf ("\n");
}

static void track_test (char * dest, int bits, int packet_size)
{
  int rate = track_rate (dest, bits, packet_size);
  printf ("track (%02x %02x %02x %02x, %d, %d) => ",
          dest [0] & 0xff, dest [1] & 0xff, dest [2] & 0xff, dest [3] & 0xff,
          bits, packet_size);
  print_fraction (rate, NULL);
  printf ("\n");
}

int main (int argc, char ** argv)
{
  /* computes priority as a fraction of MAX_PRIORITY.  For example, a
   * priority of 3/4 is MAX_PRIORITY / 4 * 3
   */
  /*
  extern int compute_priority (int is_local, int size, int dest_bits,
                               int hops_already, int hops_left,
                               int social_distance, int rate_fraction);
   */
  /* local_priority should always be 1 */
  int p = compute_priority (1, 1500, 0, 5, 10, 3, MAX_FRACTION);
  priority_test (p, SEVEN_EIGHTS, "local");

  /* examples of priority from a paper I plan to submit */
  p = compute_priority (0, 1000, 96, 1, 1, 1, 0);
  priority_test (p, THREE_QUARTERS, "friend");
  p = compute_priority (0, 1000, 96, 1, 1, 2, 0);
  priority_test (p, ONE_HALF, "1+1 hop for f2");

  int one_third   = MAX_PRIORITY / 3;
  int one_sixth   = multiply (ONE_HALF,    one_third);
  int one_twelfth = multiply (ONE_QUARTER, one_third);

  p = compute_priority (0, 500, 96, 1, 4, 2, 0);    /* first hop */
  priority_test (p, ONE_EIGHT, "1+4 hops for f2");
  p = compute_priority (0, 500, 96, 2, 3, 2, 0);    /* second hop */
  priority_test (p, one_twelfth, "2+3 hops for f2");
  p = compute_priority (0, 500, 96, 3, 2, 2, 0);    /* third hop */
  priority_test (p, one_twelfth, "3+2 hops for f2");
  p = compute_priority (0, 500, 96, 4, 1, 2, 0);    /* last hop */
  priority_test (p, ONE_EIGHT, "4+1 hops for f2");

  p = compute_priority (0, 500, 96, 1, 1, 2, one_third);    /* 1/3 rate */
  priority_test (p, one_third, "rate = 1/3");

  observe_priority (0, 500, 4, 4, 2, 3, ONE_QUARTER, "complicated");

  observe_priority (0, 500, 4, 4, 2, 4, ONE_QUARTER, "less-var social");
  observe_priority (0, 500, 4, 4, 2, 2, ONE_QUARTER, "plus-var social");

  observe_priority (0, 500, 4, 5, 2, 3, ONE_QUARTER, "less-var hops1");
  observe_priority (0, 500, 4, 3, 2, 3, ONE_QUARTER, "plus-var hops1");
  observe_priority (0, 500, 4, 4, 3, 3, ONE_QUARTER, "less-var hops2");
  observe_priority (0, 500, 4, 4, 1, 3, ONE_QUARTER, "plus-var hops2");

  observe_priority (0, 500, 3, 4, 2, 3, ONE_QUARTER, "less-var bits");
  observe_priority (0, 500, 5, 4, 2, 3, ONE_QUARTER, "plus-var bits");

  int point_three = (MAX_PRIORITY / 10) * 3;
  int point_two   = (MAX_PRIORITY / 10) * 2;
  observe_priority (0, 500, 4, 4, 2, 3, point_three, "less-var rate");
  observe_priority (0, 500, 4, 4, 2, 3, point_two  , "plus-var rate");

  char p1 [] = "hello world";
  char p2 [] = "foo bar baz qux";
  record_test (p1, strlen (p1), 0);
  record_test (p1, sizeof (p1), 0);  /* sizeof is one longer than strlen */
  record_test (p1, strlen (p1), 1);
  record_test (p2, strlen (p2), 0);
  record_test (p1, strlen (p1), 1);
  record_test (p1, sizeof (p1), 1);
  record_test (p2, strlen (p2), 1);

  char d1 [DESTINATION_SIZE];
  char d2 [DESTINATION_SIZE];
  init_dest (d1, time (NULL)      , DESTINATION_SIZE);
  init_dest (d2, time (NULL) + 123, DESTINATION_SIZE);
  track_test (d1, DESTINATION_SIZE * 8, 1000);
  track_test (d1, DESTINATION_SIZE * 8, 1500);
  track_test (d2, 10, 1000);
  track_test (d1, 10, 1500);

  return 0;
}
