/* track.c: keep track of recently received packets */

#include <stdio.h>
#include <string.h>

#include "packet.h"
#include "priority.h"

#define SAVED_DESTINATIONS	128

struct rate_record {
  char dest [DESTINATION_SIZE];
  unsigned char num_bits;
  int packet_size;
};

static struct rate_record record [SAVED_DESTINATIONS];

static int next = -1;

/* return 1 if the destination and record match, 0 otherwise */
static int matching (char * destination, int dest_bits, 
                     struct rate_record * record)
{
  int min_bits = (dest_bits < record->num_bits) ? dest_bits : record->num_bits;
  int full_bytes = min_bits / 8;
  int odd_bits = min_bits % 8;
  if ((full_bytes > 0) &&
      (memcmp (destination, record->dest, full_bytes) != 0))
    return 0;   /* no match in the full bytes */
  if (odd_bits > 0) {
    int c1 = (destination  [full_bytes] >> (8 - odd_bits)) & 0xff;
    int c2 = (record->dest [full_bytes] >> (8 - odd_bits)) & 0xff;
    if (c1 != c2)
      return 0;
  }
  return 1;  /* matches the full bytes and the odd bits if any */
}

/* record that we are sending this packet of given size to this destination */
/* return an integer, as a fraction of MAX_PRIORITY, to indicate what
 * fraction of the available bandwidth this destination is using.
 * MAX_PRIORITY is defined in priority.h
 */
int track_rate (char * destination, int dest_bits, int packet_size)
{
  if (packet_size <= 0) {
    printf ("error in track_rate: illegal packet size %d, returning epsilon\n",
            packet_size);
    return 1;
  }

  int i;
  if (next < 0) {
    /* initialize */
    for (i = 0; i < SAVED_DESTINATIONS; i++) {
      memset (record [i].dest, 0, DESTINATION_SIZE);
      record [i].num_bits = 0;
      record [i].packet_size = 0;
    }
    next = 0;
  }

/*
  for (i = 0; i < SAVED_DESTINATIONS; i++)
    printf ("track_rate %d: %02x %02x %02x %02x, %d, %d\n", i,
            record [i].dest [0] & 0xff, record [i].dest [1] & 0xff,
            record [i].dest [2] & 0xff, record [i].dest [3] & 0xff,
            record [i].num_bits, record [i].packet_size);
*/

  /* how many saved packets have the same destination as this one? */
  int matches = 0;
  int total = 0;
  for (i = 0; i < SAVED_DESTINATIONS; i++) {
    if (record [i].packet_size > 0) {
      total += record [i].packet_size;
      if (matching (destination, dest_bits, record + i))
        matches += record [i].packet_size;
    }
  }

  /* save this packet */
  memcpy (record [next].dest, destination, DESTINATION_SIZE);
  record [next].num_bits = dest_bits;
  record [next].packet_size = packet_size;
  next = (next + 1) % SAVED_DESTINATIONS;

  matches += packet_size;    /* add in this packet */
  total += packet_size;    /* add in this packet */

  if (total == 0) {
    printf ("error in track_rate: illegal total size %d, returning one\n",
            total);
    return MAX_PRIORITY;
  }
  printf ("total %d, matching %d\n", total, matches);
  return (MAX_PRIORITY / total) * matches;
}

