/* util.c: a place for useful functions used by different programs */

#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <netinet/in.h>
#include <netpacket/packet.h>
#include <arpa/inet.h>

#include "util.h"

/* print up to max of the count characters in the buffer.
 * desc is printed first unless it is null
 * a newline is printed after if print_eol
 */
void print_buffer (const char * buffer, int count, char * desc,
                   int max, int print_eol)
{
  int i;
  if (desc != NULL)
    printf ("%s (%d bytes):", desc, count);
  else
    printf ("%d bytes:", count);
  for (i = 0; i < count && i < max; i++)
    printf (" %02x", buffer [i] & 0xff);
  if (i < count)
    printf (" ...");
  if (print_eol)
    printf ("\n");
}

void print_sockaddr (struct sockaddr * sap, int addr_size)
{
  if (sap == NULL) {
    printf ("(null)");
    return;
  }
  struct sockaddr_in  * sin  = (struct sockaddr_in  *) sap;
  struct sockaddr_in6 * sin6 = (struct sockaddr_in6 *) sap;
  struct sockaddr_un  * sun  = (struct sockaddr_un  *) sap;
  struct sockaddr_ll  * sll  = (struct sockaddr_ll  *) sap;
  /* char str [INET_ADDRSTRLEN]; */
  int num_initial_zeros = 0;  /* for printing ipv6 addrs */
  int i;
  switch (sap->sa_family) {
  case AF_INET:
    printf ("ip4 %s %d/%x", inet_ntoa (sin->sin_addr), ntohs (sin->sin_port),
            ntohs (sin->sin_port));
    if ((addr_size != 0) && (addr_size != sizeof (struct sockaddr_in)))
      printf (" (size %d rather than %ld)",
              addr_size, sizeof (struct sockaddr_in));
    break;
  case AF_INET6:
    /* inet_ntop (AF_INET6, sap, str, sizeof (str)); */
    printf ("ip6 ");
    for (i = 0; i + 1 < sizeof (sin6->sin6_addr); i++)
      if ((sin6->sin6_addr.s6_addr [i] & 0xff) == 0)
        num_initial_zeros++;
      else
        break;
    if (num_initial_zeros > 0)
      printf ("::");
    for (i = num_initial_zeros; i + 1 < sizeof (sin6->sin6_addr); i++)
      printf ("%x:", sin6->sin6_addr.s6_addr [i] & 0xff);
    /* last one is not followed by : */
    printf ("%x %d/%x", sin6->sin6_addr.s6_addr [i] & 0xff,
            ntohs (sin6->sin6_port), ntohs (sin6->sin6_port));
    if ((addr_size != 0) && (addr_size != sizeof (struct sockaddr_in6)))
      printf (" (size %d rather than %ld)",
              addr_size, sizeof (struct sockaddr_in6));
    break;
  case AF_UNIX:
    printf ("unix %s", sun->sun_path);
    if ((addr_size != 0) && (addr_size != sizeof (struct sockaddr_un)))
      printf (" (size %d rather than %ld)",
              addr_size, sizeof (struct sockaddr_un));
    break;
  case AF_PACKET:
    printf ("packet protocol 0x%x if %d ha %d pkt %d address (%d)",
            sll->sll_protocol, sll->sll_ifindex, sll->sll_hatype,
            sll->sll_pkttype, sll->sll_halen);
    for (i = 0; i < sll->sll_halen; i++)
      printf (" %02x", sll->sll_addr [i] & 0xff);
    if ((addr_size != 0) && (addr_size != sizeof (struct sockaddr_ll)))
      printf (" (size %d rather than %ld)",
              addr_size, sizeof (struct sockaddr_ll));
    break;
  default:
    printf ("unknown address family %d", sap->sa_family);
    break;
  }
}

/* return 1 if the first nbits of x match the first nbits of y, else 0 */
/* where nbits is the lesser of xbits and ybits */
int matches (unsigned char * x, int xbits, unsigned char * y, int ybits)
{
  int nbits = xbits;
  if (nbits > ybits)
    nbits = ybits;
  int bytes = nbits / 8;  /* rounded-down number of bytes */
  printf ("matching %d bits, %d bytes, of ", nbits, bytes);
  print_buffer (x, bytes + 1, NULL, 6, 0);
  printf (", ");
  print_buffer (y, bytes + 1, NULL, 6, 1);
  int i;
  for (i = 0; i < bytes; i++)
    if (x [i] != y [i])
      return 0;
  if ((nbits % 8) == 0)   /* identical */
    return 1;
  int shift = 8 - nbits % 8;
  return ((((x [bytes]) & 0xff) >> shift) ==
          (((y [bytes]) & 0xff) >> shift));
}

/* useful time functions */
unsigned long long delta_us (struct timeval * t1, struct timeval * t2)
{
  if ((t1->tv_sec < t2->tv_sec) ||
      ((t1->tv_sec == t2->tv_sec) &&
       (t1->tv_usec < t2->tv_usec)))  /* t1 before t2, return 0 */
    return 0LL;
  unsigned long long result = t1->tv_usec - t2->tv_usec;
  result += (t1->tv_sec - t2->tv_sec) * US_PER_S;
  return result;
}

void add_us (struct timeval * t, unsigned long long us)
{
  t->tv_usec += us % US_PER_S;         /* add microseconds to tv_usec */
  t->tv_sec += t->tv_usec / US_PER_S;  /* any carry goes into tv_sec */
  t->tv_usec = t->tv_usec % US_PER_S;  /* tv_usec should be < 1,000,000 */
  t->tv_sec += us / US_PER_S;          /* whole seconds added to tv_sec */
}

/* set result to a random time between start + min and start + max */
void set_time_random (struct timeval * start, unsigned long long min,
                      unsigned long long max, struct timeval * result)
{
  unsigned long long delta = max - min;
  unsigned long long r = random ();
  unsigned long long us = min + r % delta;
  *result = *start;
  add_us (result, us);
}

/* if malloc is not successful, exit after printing */
void * malloc_or_fail (int bytes, char * desc)
{
  void * result = malloc (bytes);
  if (result == NULL) {
    printf ("unable to allocate %d bytes for %s\n", bytes, desc);
    exit (1);
  }
  return result;
}

/* place the values 0..n-1 at random within the given array */
void random_permute_array (int n, int * array)
{
  int i;
  for (i = 0; i < n; i++)
    array [i] = i;
  if (n <= 1)  /* done */
    return;
  /* now assign to each element a random selection of the other elements */
  for (i = 0; i < n; i++) {
    int r = random () % n;
    int swap = array [i];   /* this code works even if r == i */
    array [i] = array [r];
    array [r] = swap;
  }
  printf ("permutation of %d is", n);
  for (i = 0; i < n; i++)
    printf (" %d", array [i]);
  printf ("\n");
}


/* malloc and return an n-element int array containing the values 0..n-1
 * in some random permuted order */
int * random_permute (int n)
{
  int * result = malloc_or_fail (n * sizeof (int), "random_permute");
  random_permute_array (n, result);
  return result;
}
